"""Quantum state distinguishability scenarios and protocols."""
from typing import List

import cvxpy
import numpy as np

from toqito.channels import partial_transpose


class StateDistinguish:
    r"""
    Class to distinguish ensembles of quantum states.

    Test.
    """

    def __init__(self, states: List[np.ndarray], probs: List[float] = None):
        """Create distinguishability scenarios.

        :param states: A list of density operators (matrices) corresponding to
                       quantum states.
        :param probs: A list of probabilities where `probs[i]` corresponds to
                      the probability that `states[i]` is selected by Alice.
        """
        # Assume that at least one state is provided.
        if states is None or states == []:
            raise ValueError(
                "InvalidStates: There must be at least one state provided."
            )

        # Assume uniform probability if no specific distribution is given.
        if probs is None:
            probs = [1 / len(states)] * len(states)
        if not np.isclose(sum(probs), 1):
            raise ValueError("Invalid: Probabilities must sum to 1.")

        _, dim_y = states[0].shape

        # The variable `states` is provided as a list of vectors. Transform them
        # into density matrices.
        if dim_y == 1:
            for i, state_ket in enumerate(states):
                states[i] = state_ket * state_ket.conj().T

        self._states = states
        self._probs = probs

    def state_distinguishability(self) -> float:
        r"""
        Compute probability of state distinguishability [ELD03]_.

        The "quantum state distinguishability" problem involves a collection of
        :math:`n` quantum states

        .. math::
            \rho = \{ \rho_0, \ldots, \rho_n \},

        as well as a list of corresponding probabilities

        .. math::
            p = \{ p_0, \ldots, p_n \}

        Alice chooses :math:`i` with probability :math:`p_i` and creates the state
        :math:`rho_i`

        Bob wants to guess which state he was given from the collection of states.

        This function implements the following semidefinite program that provides
        the optimal probability with which Bob can conduct quantum state
        distinguishability.

        .. math::

            \begin{align*}
                \text{maximize:} \quad & \sum_{i=0}^n p_i \langle M_i,
                \rho_i \rangle \\
                \text{subject to:} \quad & M_0 + \ldots + M_n = \mathbb{I},\\
                                         & M_0, \ldots, M_n \geq 0
            \end{align*}

        Examples
        ==========

        State distinguishability for two state density matrices.

        >>> from toqito.states import basis, bell
        >>> from toqito.state_distinguish import StateDistinguish
        >>> e_0, e_1 = basis(2, 0), basis(2, 1)
        >>> e_00 = e_0 * e_0.conj().T
        >>> e_11 = e_1 * e_1.conj().T
        >>> states = [e_00, e_11]
        >>> probs = [1 / 2, 1 / 2]
        >>> s_d = StateDistinguish(states, probs)
        >>> res = s_d.state_distinguishability(states, probs)
        0.5000000000006083

        References
        ==========
        .. [ELD03] Eldar, Yonina C.
            "A semidefinite programming approach to optimal unambiguous
            discrimination of quantum states."
            IEEE Transactions on information theory 49.2 (2003): 446-456.
            https://arxiv.org/abs/quant-ph/0206093


        :return: The optimal probability with which Bob can distinguish the state.
        """
        obj_func = []
        measurements = []
        constraints = []

        dim_x, _ = self._states[0].shape
        for i, _ in enumerate(self._states):
            measurements.append(cvxpy.Variable((dim_x, dim_x), PSD=True))

            obj_func.append(
                self._probs[i] * cvxpy.trace(self._states[i].conj().T @ measurements[i])
            )

        constraints.append(sum(measurements) == np.identity(dim_x))

        objective = cvxpy.Maximize(sum(obj_func))
        problem = cvxpy.Problem(objective, constraints)
        sol_default = problem.solve()

        return 1 / len(self._states) * sol_default

    def ppt_distinguishability(self) -> float:
        r"""
        Compute probability of distinguishing a state via PPT measurements [COS13]_.

        Implements the semidefinite program (SDP) whose optimal value is equal to
        the maximum probability of perfectly distinguishing orthogonal maximally
        entangled states using any PPT measurement; a measurement whose operators
        are positive under partial transpose. This SDP was explicitly provided in
        [COS13]_.

        Specifically, the function implements the dual problem (as this is
        computationally more efficient) and is defined as:

        .. math::

            \begin{equation}
                \begin{aligned}
                    \text{minimize:} \quad & \frac{1}{k} \text{Tr}(Y) \\
                    \text{subject to:} \quad & Y \geq \text{T}_{\mathcal{A}}
                                              (\rho_j), \quad j = 1, \ldots, k, \\
                                             & Y \in \text{Herm}(\mathcal{A} \otimes
                                              \mathcal{B}).
                \end{aligned}
            \end{equation}

        Examples
        ==========

        Consider the following Bell states

        .. math::
            \begin{equation}
                \begin{aligned}
                |\psi_0 \rangle = \frac{|00\rangle + |11\rangle}{\sqrt{2}}, &\quad
                |\psi_1 \rangle = \frac{|01\rangle + |10\rangle}{\sqrt{2}}, \\
                |\psi_2 \rangle = \frac{|01\rangle - |10\rangle}{\sqrt{2}}, &\quad
                |\psi_3 \rangle = \frac{|00\rangle - |11\rangle}{\sqrt{2}}.
                \end{aligned}
            \end{equation}

        It was illustrated in [YDY12]_ that for the following set of states:

        The PPT distinguishability of the following states

        .. math::
            \begin{equation}
                \rho_1^{(2)} = \psi_0 \otimes \psi_0, \quad
                \rho_2^{(2)} = \psi_1 \otimes \psi_1, \quad
            \end{equation}

        should yield :math:`7/8 ~ 0.875` as was proved in [YDY12]_.

        >>> from toqito.states import bell
        >>> from toqito.state_distinguish import StateDistinguish
        >>> # Bell vectors:
        >>> psi_0 = bell(0)
        >>> psi_1 = bell(2)
        >>> psi_2 = bell(3)
        >>> psi_3 = bell(1)
        >>>
        >>> # YYD vectors from [YDY12]_.
        >>> x_1 = np.kron(psi_0, psi_0)
        >>> x_2 = np.kron(psi_1, psi_3)
        >>> x_3 = np.kron(psi_2, psi_3)
        >>> x_4 = np.kron(psi_3, psi_3)
        >>>
        >>> # YYD density matrices.
        >>> rho_1 = x_1 * x_1.conj().T
        >>> rho_2 = x_2 * x_2.conj().T
        >>> rho_3 = x_3 * x_3.conj().T
        >>> rho_4 = x_4 * x_4.conj().T
        >>>
        >>> states = [rho_1, rho_2, rho_3, rho_4]
        >>> probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]
        >>> ydy = StateDistinguish(states, probs)
        >>> ydy.ppt_distinguishability()
        0.875

        References
        ==========
        .. [COS13] Cosentino, Alessandro.
            "Positive-partial-transpose-indistinguishable states via semidefinite
            programming."
            Physical Review A 87.1 (2013): 012321.
            https://arxiv.org/abs/1205.1031

        .. [YDY12] Yu, Nengkun, Runyao Duan, and Mingsheng Ying.
            "Four locally indistinguishable ququad-ququad orthogonal
            maximally entangled states."
            Physical review letters 109.2 (2012): 020506.
            https://arxiv.org/abs/1107.3224

        :return: The optimal probability with which the states can be distinguished
                 via PPT measurements.
        """
        constraints = []
        dim_x, _ = self._states[0].shape
        y_var = cvxpy.Variable((dim_x, dim_x), hermitian=True)
        objective = (
            1 / len(self._states) * cvxpy.Minimize(cvxpy.trace(cvxpy.real(y_var)))
        )

        dim = int(np.log2(dim_x))
        dim_list = [2] * int(np.log2(dim_x))
        sys_list = list(range(1, dim, 2))

        for i, _ in enumerate(self._states):
            constraints.append(
                cvxpy.real(y_var)
                >> partial_transpose(self._states[i], sys=sys_list, dim=dim_list)
            )

        problem = cvxpy.Problem(objective, constraints)
        sol_default = problem.solve()

        return sol_default

    def conclusive_state_exclusion(self) -> float:
        r"""
        Compute probability of conclusive single state exclusion.

        The "quantum state exclusion" problem involves a collection of :math:`n`
        quantum states

        .. math::
            \rho = \{ \rho_0, \ldots, \rho_n \},

        as well as a list of corresponding probabilities

        .. math::
            p = \{ p_0, \ldots, p_n \}

        Alice chooses :math:`i` with probability :math:`p_i` and creates the state
        :math:`\rho_i`.

        Bob wants to guess which state he was *not* given from the collection of
        states. State exclusion implies that ability to discard (with certainty) at
        least one out of the "n" possible quantum states by applying a measurement.

        This function implements the following semidefinite program that provides
        the optimal probability with which Bob can conduct quantum state exclusion.

            .. math::
                \begin{equation}
                    \begin{aligned}
                        \text{minimize:} \quad & \sum_{i=0}^n p_i \langle M_i,
                                                    \rho_i \rangle \\
                        \text{subject to:} \quad & M_0 + \ldots + M_n =
                                                   \mathbb{I}, \\
                                                 & M_0, \ldots, M_n >= 0
                    \end{aligned}
                \end{equation}

        The conclusive state exclusion SDP is written explicitly in [BJOP14]_. The
        problem of conclusive state exclusion was also thought about under a
        different guise in [PBR12]_.

        Examples
        ==========

        Consider the following two Bell states

        .. math::
            u_0 = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \\
            u_1 = \frac{1}{\sqrt{2}} \left( |00 \rangle - |11 \rangle \right).

        For the corresponding density matrices :math:`\rho_0 = u_0 u_0^*` and
        :math:`\rho_1 = u_1 u_1^*`, we may construct a set

        .. math::
            \rho = \{\rho_0, \rho_1 \}

        such that

        .. math::
            p = \{1/2, 1/2\}.

        It is not possible to conclusively exclude either of the two states. We can
        see that the result of the function in `toqito` yields a value of :math`0`
        as the probability for this to occur.

        >>> from toqito.state_distinguish import StateDistinguish
        >>> from toqito.states import bell
        >>> import numpy as np
        >>> rho1 = bell(0) * bell(0).conj().T
        >>> rho2 = bell(1) * bell(1).conj().T
        >>>
        >>> states = [rho1, rho2]
        >>> probs = [1/2, 1/2]
        >>>
        >>> s_d = StateDistinguish(states, probs)
        >>> s_d.conclusive_state_exclusion()
        1.6824720366950206e-09

        References
        ==========
        .. [PBR12] "On the reality of the quantum state"
            Pusey, Matthew F., Jonathan Barrett, and Terry Rudolph.
            Nature Physics 8.6 (2012): 475-478.
            arXiv:1111.3328

        .. [BJOP14] "Conclusive exclusion of quantum states"
            Somshubhro Bandyopadhyay, Rahul Jain, Jonathan Oppenheim,
            Christopher Perry
            Physical Review A 89.2 (2014): 022336.
            arXiv:1306.4683

        :return: The optimal probability with which Bob can guess the state he was
                 not given from `states`.
        """
        obj_func = []
        measurements = []
        constraints = []
        dim_x, _ = self._states[0].shape

        for i, _ in enumerate(self._states):
            measurements.append(cvxpy.Variable((dim_x, dim_x), PSD=True))

            obj_func.append(
                self._probs[i] * cvxpy.trace(self._states[i].conj().T @ measurements[i])
            )

        constraints.append(sum(measurements) == np.identity(dim_x))

        if np.iscomplexobj(self._states[0]):
            objective = cvxpy.Minimize(cvxpy.real(sum(obj_func)))
        else:
            objective = cvxpy.Minimize(sum(obj_func))

        problem = cvxpy.Problem(objective, constraints)
        sol_default = problem.solve()

        return 1 / len(self._states) * sol_default

    def unambiguous_state_exclusion(self) -> float:
        r"""
        Compute probability of unambiguous state exclusion [BJOPUS14]_.

        This function implements the following semidefinite program that provides
        the optimal probability with which Bob can conduct quantum state exclusion.

        .. math::

            \begin{align*}
                \text{maximize:} \quad & \sum_{i=0}^n \sum_{j=0}^n
                                         \langle M_i, \rho_j \rangle \\
                \text{subject to:} \quad & \sum_{i=0}^n M_i \leq \mathbb{I},\\
                                         & \text{Tr}(\rho_i M_i) = 0,
                                           \quad \quad \forall 1  \leq i \leq n, \\
                                         & M_0, \ldots, M_n \geq 0
            \end{align*}

        Examples
        ==========

        Consider the following two Bell states

        .. math::
            u_0 = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \\
            u_1 = \frac{1}{\sqrt{2}} \left( |00 \rangle - |11 \rangle \right).

        For the corresponding density matrices :math:`\rho_0 = u_0 u_0^*` and
        :math:`\rho_1 = u_1 u_1^*`, we may construct a set

        .. math::
            \rho = \{\rho_0, \rho_1 \}

        such that

        .. math::
            p = \{1/2, 1/2\}.

        It is not possible to unambiguously exclude either of the two states. We can
        see that the result of the function in `toqito` yields a value of :math:`0`
        as the probability for this to occur.

        >>> from toqito.state_distinguish import StateDistinguish
        >>> from toqito.states import bell
        >>> import numpy as np
        >>> rho1 = bell(0) * bell(0).conj().T
        >>> rho2 = bell(1) * bell(1).conj().T
        >>>
        >>> states = [rho1, rho2]
        >>> probs = [1/2, 1/2]
        >>>
        >>> s_d = StateDistinguish(states, probs)
        >>> s_d.unambiguous_state_exclusion()
        -7.250173600116328e-18

        References
        ==========
        .. [BJOPUS14] "Conclusive exclusion of quantum states"
            Somshubhro Bandyopadhyay, Rahul Jain, Jonathan Oppenheim,
            Christopher Perry
            Physical Review A 89.2 (2014): 022336.
            arXiv:1306.4683

        :return: The optimal probability with which Bob can guess the state he was
                 not given from `states` with certainty.
        """
        obj_func = []
        measurements = []
        constraints = []
        dim_x, _ = self._states[0].shape

        for i, _ in enumerate(self._states):
            measurements.append(cvxpy.Variable((dim_x, dim_x), PSD=True))

            obj_func.append(
                self._probs[i] * cvxpy.trace(self._states[i].conj().T @ measurements[i])
            )

            constraints.append(cvxpy.trace(self._states[i] @ measurements[i]) == 0)

        constraints.append(sum(measurements) <= np.identity(dim_x))

        if np.iscomplexobj(self._states[0]):
            objective = cvxpy.Maximize(cvxpy.real(sum(obj_func)))
        else:
            objective = cvxpy.Maximize(sum(obj_func))

        problem = cvxpy.Problem(objective, constraints)
        sol_default = problem.solve()

        return 1 / len(self._states) * sol_default



"""Tests for is_diagonal function."""
import unittest
import numpy as np

from toqito.matrix_props import is_diagonal


class TestIsDiagonal(unittest.TestCase):

    """Unit test for is_diagonal."""

    def test_is_diagonal(self):
        """Test if matrix is diagonal."""
        mat = np.array([[1, 0, 0], [0, 2, 0], [0, 0, 3]])
        self.assertEqual(is_diagonal(mat), True)

    def test_is_non_diagonal(self):
        """Test non-diagonal matrix."""
        mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.assertEqual(is_diagonal(mat), False)

    def test_non_square(self):
        """Test on a non-square matrix."""
        mat = np.array([[1, 0, 0], [0, 1, 0]])
        self.assertEqual(is_diagonal(mat), False)


if __name__ == "__main__":
    unittest.main()



"""Tests for random_density_matrix function."""
import unittest

from toqito.random import random_density_matrix
from toqito.matrix_props import is_density


class TestRandomDensity(unittest.TestCase):

    """Unit test for random_unitary."""

    def test_random_density_not_real(self):
        """Generate random non-real density matrix."""
        mat = random_density_matrix(2)
        self.assertEqual(is_density(mat), True)

    def test_random_density_real(self):
        """Generate random real density matrix."""
        mat = random_density_matrix(2, True)
        self.assertEqual(is_density(mat), True)

    def test_random_density_not_real_bures(self):
        """Random non-real density matrix according to Bures metric."""
        mat = random_density_matrix(2, distance_metric="bures")
        self.assertEqual(is_density(mat), True)

    def test_random_density_not_real_k_param(self):
        """Generate random non-real density matrix wih k_param."""
        mat = random_density_matrix(2, distance_metric="bures")
        self.assertEqual(is_density(mat), True)

    def test_random_density_not_real_all_params(self):
        """Generate random non-real density matrix all params."""
        mat = random_density_matrix(2, True, 2, "haar")
        self.assertEqual(is_density(mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for chessboard function."""
import unittest
import numpy as np

from toqito.states import chessboard


class TestChessboard(unittest.TestCase):

    """Unit test for chessboard."""

    def test_chessboard(self):
        """The chessboard_state."""
        res = chessboard([1, 2, 3, 4, 5, 6], 7, 8)
        self.assertEqual(np.isclose(res[0][0], 0.22592592592592592), True)

    def test_chessboard_default_s(self):
        """The chessboard_state with default `s_param`."""
        res = chessboard([1, 2, 3, 4, 5, 6], 7)
        self.assertEqual(np.isclose(res[0][0], 0.29519938056523426), True)

    def test_chessboard_default_s_t(self):
        """The chessboard_state with default `s_param` and `t_param`."""
        res = chessboard([1, 2, 3, 4, 5, 6])
        self.assertEqual(np.isclose(res[0][0], 0.3863449236810438), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for is_density function."""
import unittest
import numpy as np

from toqito.matrix_props import is_density
from toqito.random import random_density_matrix


class TestIsDensity(unittest.TestCase):

    """Unit test for is_density."""

    def test_is_density_ndarray(self):
        """Test if ndarray type is density matrix."""
        mat = random_density_matrix(2, True)
        self.assertEqual(is_density(mat), True)

    def test_is_density_npmatrix(self):
        """Test if np.matrix type is density matrix."""
        mat = np.matrix(random_density_matrix(4))
        self.assertEqual(is_density(mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for QuantumMoney class."""
import unittest
import numpy as np

from toqito.states import basis
from toqito.state_ops import tensor
from toqito.nonlocal_games.quantum_money import QuantumMoney


class TestQuantumMoney(unittest.TestCase):

    """Unit test for counterfeit_attack."""

    def test_counterfeit_attack_wiesner_money(self):
        """Probability of counterfeit attack on Wiesner's quantum money."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        e_p = (e_0 + e_1) / np.sqrt(2)
        e_m = (e_0 - e_1) / np.sqrt(2)

        e_000 = tensor(e_0, e_0, e_0)
        e_111 = tensor(e_1, e_1, e_1)
        e_ppp = tensor(e_p, e_p, e_p)
        e_mmm = tensor(e_m, e_m, e_m)

        q_a = (
            1
            / 4
            * (
                e_000 * e_000.conj().T
                + e_111 * e_111.conj().T
                + e_ppp * e_ppp.conj().T
                + e_mmm * e_mmm.conj().T
            )
        )
        wiesner = QuantumMoney(q_a)
        res = wiesner.counterfeit_attack()
        self.assertEqual(np.isclose(res, 3 / 4), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for random_povm function."""
import unittest
import numpy as np

from toqito.random import random_povm


class TestRandomPOVM(unittest.TestCase):

    """Unit test for random_povm."""

    def test_random_unitary_not_real(self):
        """Generate random POVMs and check that they sum to the identity."""
        dim, num_inputs, num_outputs = 2, 2, 2
        povms = random_povm(dim, num_inputs, num_outputs)
        self.assertEqual(
            np.allclose(povms[:, :, 0, 0] + povms[:, :, 0, 1], np.identity(dim)), True
        )


if __name__ == "__main__":
    unittest.main()



"""Tests for is_normal function."""
import unittest
import numpy as np

from toqito.matrix_props import is_normal


class TestIsNormal(unittest.TestCase):

    """Unit test for is_normal."""

    def test_is_normal(self):
        """Test that normal matrix returns True."""
        mat = np.identity(4)
        self.assertEqual(is_normal(mat), True)

    def test_is_not_normal(self):
        """Test that non-normal matrix returns False."""
        mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        self.assertEqual(is_normal(mat), False)


if __name__ == "__main__":
    unittest.main()



"""Tests for symmetric_projection function."""
import unittest
import numpy as np

from toqito.perms import symmetric_projection


class TestSymmetricProjection(unittest.TestCase):

    """Unit test for symmetric_projection."""

    def test_symmetric_projection_dim_2_pval_1(self):
        """Symmetric_projection where the dimension is 2 and p_val is 1."""
        res = symmetric_projection(dim=2, p_val=1).todense()
        expected_res = np.array([[1, 0], [0, 1]])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_symmetric_projection_2(self):
        """Generates the symmetric_projection where the dimension is 2."""
        res = symmetric_projection(dim=2).todense()
        expected_res = np.array(
            [[1, 0, 0, 0], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0, 0, 0, 1]]
        )

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_symmetric_projection_dim_2_partial_true(self):
        """Symmetric_projection where the dimension is 2 and partial is True."""
        res = symmetric_projection(dim=2, p_val=2, partial=True).todense()
        expected_res = np.array(
            [[0, 0, 1], [-1 / np.sqrt(2), 0, 0], [-1 / np.sqrt(2), 0, 0], [0, 1, 0]]
        )

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for hedging_value function."""
import unittest
from numpy import kron, cos, sin, pi, sqrt, isclose

from toqito.states import basis
from toqito.nonlocal_games.quantum_hedging import QuantumHedging


class TestQuantumHedging(unittest.TestCase):

    """Unit test for hedging_value."""

    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00, e_01 = kron(e_0, e_0), kron(e_0, e_1)
    e_10, e_11 = kron(e_1, e_0), kron(e_1, e_1)

    alpha = 1 / sqrt(2)
    theta = pi / 8

    w_var = alpha * cos(theta) * e_00 + sqrt(1 - alpha ** 2) * sin(theta) * e_11

    l_1 = -alpha * sin(theta) * e_00 + sqrt(1 - alpha ** 2) * cos(theta) * e_11

    l_2 = alpha * sin(theta) * e_10

    l_3 = sqrt(1 - alpha ** 2) * cos(theta) * e_01

    q_1 = w_var * w_var.conj().T
    q_0 = l_1 * l_1.conj().T + l_2 * l_2.conj().T + l_3 * l_3.conj().T

    def test_max_prob_outcome_a_primal_1_dim(self):
        """
        Maximal probability of outcome "a" when dim == 1.

        The primal problem of the hedging semidefinite program.
        """
        q_0 = TestQuantumHedging.q_0
        hedging_value = QuantumHedging(q_0, 1)
        self.assertEqual(
            isclose(hedging_value.max_prob_outcome_a_primal(), cos(pi / 8) ** 2), True
        )

    def test_max_prob_outcome_a_primal_2_dim(self):
        """
        Test maximal probability of outcome "a" when dim == 2.

        The primal problem of the hedging semidefinite program.
        """
        q_00 = kron(TestQuantumHedging.q_0, TestQuantumHedging.q_0)
        hedging_value = QuantumHedging(q_00, 2)
        self.assertEqual(
            isclose(hedging_value.max_prob_outcome_a_primal(), cos(pi / 8) ** 4), True
        )

    def test_max_prob_outcome_a_dual_1_dim(self):
        """
        Test maximal probability of outcome "a" when dim == 1.

        The dual problem of the hedging semidefinite program.
        """
        q_0 = TestQuantumHedging.q_0
        hedging_value = QuantumHedging(q_0, 1)
        self.assertEqual(
            isclose(hedging_value.max_prob_outcome_a_dual(), cos(pi / 8) ** 2), True
        )

    def test_max_prob_outcome_a_dual_2_dim(self):
        """
        Test maximal probability of outcome "a" when dim == 2.

        The dual problem of the hedging semidefinite program.
        """
        q_00 = kron(TestQuantumHedging.q_0, TestQuantumHedging.q_0)
        hedging_value = QuantumHedging(q_00, 2)
        self.assertEqual(
            isclose(hedging_value.max_prob_outcome_a_dual(), cos(pi / 8) ** 4), True
        )

    def test_min_prob_outcome_a_primal_1_dim(self):
        """
        Test minimal probability of outcome "a" when dim == 1.

        The primal problem of the hedging semidefinite program.
        """
        q_1 = TestQuantumHedging.q_1
        hedging_value = QuantumHedging(q_1, 1)
        self.assertEqual(
            isclose(hedging_value.min_prob_outcome_a_primal(), 0, atol=0.01), True
        )

    def test_min_prob_outcome_a_primal_2_dim(self):
        """
        Test minimal probability of outcome "a" when dim == 2.

        The primal problem of the hedging semidefinite program.
        """
        q_11 = kron(TestQuantumHedging.q_1, TestQuantumHedging.q_1)
        hedging_value = QuantumHedging(q_11, 2)
        self.assertEqual(
            isclose(hedging_value.min_prob_outcome_a_primal(), 0, atol=0.01), True
        )

    def test_min_prob_outcome_a_dual_1_dim(self):
        """
        Test minimal probability of outcome "a" when dim == 1.

        The dual problem of the hedging semidefinite program.
        """
        q_1 = TestQuantumHedging.q_1
        hedging_value = QuantumHedging(q_1, 1)
        self.assertEqual(
            isclose(hedging_value.min_prob_outcome_a_dual(), 0, atol=0.01), True
        )

    def test_min_prob_outcome_a_dual_2_dim(self):
        """
        Test minimal probability of outcome "a" when dim == 2.

        The dual problem of the hedging semidefinite program.
        """
        q_11 = kron(TestQuantumHedging.q_1, TestQuantumHedging.q_1)
        hedging_value = QuantumHedging(q_11, 2)
        self.assertEqual(
            isclose(hedging_value.min_prob_outcome_a_dual(), 0, atol=0.01), True
        )


if __name__ == "__main__":
    unittest.main()



"""Tests for gen_bell function."""
import unittest
import numpy as np

from toqito.states import bell
from toqito.states import gen_bell


class TestGenBell(unittest.TestCase):

    """Unit test for gen_bell."""

    def test_gen_bell_0_0_2(self):
        """Generalized Bell state for k_1 = k_2 = 0 and dim = 2."""
        dim = 2
        k_1 = 0
        k_2 = 0

        expected_res = bell(0) * bell(0).conj().T

        res = gen_bell(k_1, k_2, dim)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gen_bell_0_1_2(self):
        """Generalized Bell state for k_1 = 0, k_2 = 1 and dim = 2."""
        dim = 2
        k_1 = 0
        k_2 = 1

        expected_res = bell(1) * bell(1).conj().T

        res = gen_bell(k_1, k_2, dim)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gen_bell_1_0_2(self):
        """Generalized Bell state for k_1 = 1, k_2 = 0 and dim = 2."""
        dim = 2
        k_1 = 1
        k_2 = 0

        expected_res = bell(2) * bell(2).conj().T

        res = gen_bell(k_1, k_2, dim)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gen_bell_1_1_2(self):
        """Generalized Bell state for k_1 = 1, k_2 = 1 and dim = 2."""
        dim = 2
        k_1 = 1
        k_2 = 1

        expected_res = bell(3) * bell(3).conj().T

        res = gen_bell(k_1, k_2, dim)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for concurrence function."""
import unittest
import numpy as np

from toqito.states import basis
from toqito.state_props import concurrence


class TestConcurrence(unittest.TestCase):

    """Unit test for concurrence."""

    def test_concurrence_entangled(self):
        """The concurrence on maximally entangled Bell state."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)

        u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
        rho = u_vec * u_vec.conj().T

        res = concurrence(rho)
        self.assertEqual(np.isclose(res, 1), True)

    def test_concurrence_separable(self):
        """The concurrence of a product state is zero."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        v_vec = np.kron(e_0, e_1)
        sigma = v_vec * v_vec.conj().T

        res = concurrence(sigma)
        self.assertEqual(np.isclose(res, 0), True)

    def test_invalid_dim(self):
        """Tests for invalid dimension inputs."""
        with self.assertRaises(ValueError):
            rho = np.identity(5)
            concurrence(rho)


if __name__ == "__main__":
    unittest.main()



"""Tests for gell_man function."""
import unittest
import numpy as np

from scipy.sparse import csr_matrix
from toqito.matrices import gell_mann


class TestGellMann(unittest.TestCase):

    """Unit test for gell_mann."""

    def test_gell_mann_idx_0(self):
        """Gell-Mann operator for index = 0."""
        expected_res = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        res = gell_mann(0)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_1(self):
        """Gell-Mann operator for index = 1."""
        expected_res = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
        res = gell_mann(1)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_2(self):
        """Gell-Mann operator for index = 2."""
        expected_res = np.array([[0, -1j, 0], [1j, 0, 0], [0, 0, 0]])
        res = gell_mann(2)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_3(self):
        """Gell-Mann operator for index = 3."""
        expected_res = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 0]])
        res = gell_mann(3)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_4(self):
        """Gell-Mann operator for index = 4."""
        expected_res = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])

        res = gell_mann(4)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_5(self):
        """Gell-Mann operator for index = 5."""
        expected_res = np.array([[0, 0, -1j], [0, 0, 0], [1j, 0, 0]])

        res = gell_mann(5)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_6(self):
        """Gell-Mann operator for index = 6."""
        expected_res = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]])

        res = gell_mann(6)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_7(self):
        """Gell-Mann operator for index = 7."""
        expected_res = np.array([[0, 0, 0], [0, 0, -1j], [0, 1j, 0]])

        res = gell_mann(7)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_idx_8(self):
        """Gell-Mann operator for index = 8."""
        expected_res = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -2]]) / np.sqrt(3)
        res = gell_mann(8)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_invalid_idx(self):
        """Invalid Gell-Mann parameters."""
        with self.assertRaises(ValueError):
            gell_mann(9)

    def test_gell_mann_sparse(self):
        """Test sparse Gell-Mann matrix."""
        res = gell_mann(3, is_sparse=True)
        self.assertEqual(isinstance(res, csr_matrix), True)


if __name__ == "__main__":
    unittest.main()



"""Properties of matrices and vectors."""
from typing import Union

import numpy as np


__all__ = [
    "is_commuting",
    "is_density",
    "is_diagonal",
    "is_hermitian",
    "is_normal",
    "is_pd",
    "is_projection",
    "is_psd",
    "is_square",
    "is_symmetric",
    "is_unitary",
]


def is_commuting(mat_1: np.ndarray, mat_2: np.ndarray) -> bool:
    r"""
    Determine if two linear operators commute with each other [WikCom]_.

    For any pair of operators :math:`X, Y \in \text{L}(\mathcal{X})`, the
    Lie bracket :math:`\left[X, Y\right] \in \text{L}(\mathcal{X})` is defined
    as

    .. math::
        \left[X, Y\right] = XY - YX.

    It holds that :math:`` if and only if :math:`X` and :math:`Y` commute
    [WatCom18]_.

    Examples
    ==========

    Consider the following matrices:

    .. math::
        A = \begin{pmatrix} 0 & 1 \\ 0 & 0 \end{pmatrix},
        \quad \text{and} \quad
        B = \begin{pmatrix} 1 & 0 \\ 0 & 0 \end{pmatrix}.

    It holds that :math:`AB=0`, however

    .. math::
        BA = \begin{pmatrix} 0 & 1 \\ 0 & 0 \end{pmatrix} = A,

    and hence, do not commute.

    >>> from toqito.matrix_props import is_commuting
    >>> import numpy as np
    >>> mat_1 = np.array([[0, 1], [0, 0]])
    >>> mat_2 = np.array([[1, 0], [0, 0]])
    >>> is_commuting(mat_1, mat_2)
    False

    Consider the following pair of matrices

    .. math::
        A = \begin{pmatrix}
            1 & 0 & 0 \\
            0 & 1 & 0 \\
            1 & 0 & 2
            \end{pmatrix}, \quad \text{and} \quad
        B = \begin{pmatrix}
            2 & 4 & 0 \\
            3 & 1 & 0 \\
            -1 & -4 & 1
            \end{pmatrix}.

    It may be verified that :math:`AB = BA = 0`, and therefore :math`A` and
    :math:`B` commute.

    >>> from toqito.matrix_props import is_commuting
    >>> import numpy as np
    >>> mat_1 = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 2]])
    >>> mat_2 = np.array([[2, 4, 0], [3, 1, 0], [-1, -4, 1]])
    >>> is_commuting(mat_1, mat_2)
    True

    References
    ==========
    .. [WikCom] Wikipedia: Commuting matrices
        https://en.wikipedia.org/wiki/Commuting_matrices

    .. [WatCom18] Watrous, John.
        "The theory of quantum information."
        Section: "Lie brackets and commutants".
        Cambridge University Press, 2018.

    :param mat_1: First matrix to check.
    :param mat_2: Second matrix to check.
    :return: Return `True` if `mat_1` commutes with `mat_2` and False otherwise.
    """
    return np.allclose(np.matmul(mat_1, mat_2) - np.matmul(mat_2, mat_1), 0)


def is_density(mat: np.ndarray) -> bool:
    r"""
    Check if matrix is a density matrix [WikDensity]_.

    A matrix is a density matrix if its trace is equal to one and it has the
    property of being positive semidefinite (PSD).

    Examples
    ==========

    Consider the Bell state

    .. math::
       u = \frac{1}{\sqrt{2}} |00 \rangle + \frac{1}{\sqrt{2}} |11 \rangle.

    Constructing the matrix :math:`\rho = u u^*` defined as

    .. math::
        \rho = \frac{1}{2} \begin{pmatrix}
                                1 & 0 & 0 & 1 \\
                                0 & 0 & 0 & 0 \\
                                0 & 0 & 0 & 0 \\
                                1 & 0 & 0 & 1
                           \end{pmatrix}

    our function indicates that this is indeed a density operator as the trace
    of :math:`\rho` is equal to :math:`1` and the matrix is positive
    semidefinite:

    >>> from toqito.matrix_props import is_density
    >>> from toqito.states import bell
    >>> import numpy as np
    >>> rho = bell(0) * bell(0).conj().T
    >>> is_density(rho)
    True

    Alternatively, the following example matrix :math:`\sigma` defined as

    .. math::
        \sigma = \frac{1}{2} \begin{pmatrix}
                                1 & 2 \\
                                3 & 1
                             \end{pmatrix}

    does satisfy :math:`\text{Tr}(\sigma) = 1`, however fails to be positive
    semidefinite, and is therefore not a density operator. This can be
    illustrated using `toqito` as follows.

    >>> from toqito.matrix_props import is_density
    >>> from toqito.states import bell
    >>> import numpy as np
    >>> sigma = 1/2 * np.array([[1, 2], [3, 1]])
    >>> is_density(sigma)
    False

    References
    ==========
    .. [WikDensity] Wikipedia: Density matrix
        https://en.wikipedia.org/wiki/Density_matrix

    :param mat: Matrix to check.
    :return: Return `True` if matrix is a density matrix, and `False`
             otherwise.
    """
    return is_psd(mat) and np.isclose(np.trace(mat), 1)


def is_diagonal(mat: np.ndarray) -> bool:
    r"""
    Determine if a matrix is diagonal [WikDiag]_.

    A matrix is diagonal if the matrix is square and if the diagonal of the
    matrix is non-zero, while the off-diagonal elements are all zero.

    The following is an example of a 3-by-3 diagonal matrix:

    .. math::
        \begin{equation}
            \begin{pmatrix}
                1 & 0 & 0 \\
                0 & 2 & 0 \\
                0 & 0 & 3
            \end{pmatrix}
        \end{equation}

    This quick implementation is given by Daniel F. from StackOverflow in
    [SODIA]_.

    Examples
    ==========

    Consider the following diagonal matrix

    .. math::
        A = \begin{pmatrix}
                1 & 0 \\
                0 & 1
            \end{pmatrix}.

    Our function indicates that this is indeed a diagonal matrix:

    >>> from toqito.matrix_props import is_diagonal
    >>> import numpy as np
    >>> A = np.array([[1, 0], [0, 1]])
    >>> is_diagonal(A)
    True

    Alternatively, the following example matrix

    .. math::
        B = \begin{pmatrix}
                1 & 2 \\
                3 & 4
            \end{pmatrix}

    is not diagonal, as shown using `toqito`

    >>> from toqito.matrix_props import is_diagonal
    >>> import numpy as np
    >>> B = np.array([[1, 2], [3, 4]])
    >>> is_diagonal(B)
    False

    References
    ==========
    .. [WikDiag] Wikipedia: Diagonal matrix
        https://en.wikipedia.org/wiki/Diagonal_matrix

    .. [SODIA] StackOverflow post
        https://stackoverflow.com/questions/43884189/

    :param mat: The matrix to check.
    :return: Returns True if the matrix is diagonal and False otherwise.
    """
    if not is_square(mat):
        return False
    i, j = mat.shape
    test = mat.reshape(-1)[:-1].reshape(i - 1, j + 1)
    return ~np.any(test[:, 1:])


def is_hermitian(mat: np.ndarray) -> bool:
    r"""
    Check if matrix is Hermitian [WikHerm]_.

    A Hermitian matrix is a complex square matrix that is equal to its own
    conjugate transpose.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                2 & 2 +1j & 4 \\
                2 - 1j & 3 & 1j \\
                4 & -1j & 1
            \end{pmatrix}

    our function indicates that this is indeed a Hermitian matrix as it holds that

    .. math::
        A = A^*.

    >>> from toqito.matrix_props import is_hermitian
    >>> import numpy as np
    >>> mat = np.array([[2, 2 + 1j, 4], [2 - 1j, 3, 1j], [4, -1j, 1]])
    >>> is_hermitian(mat)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    is not Hermitian.

    >>> from toqito.matrix_props import is_hermitian
    >>> import numpy as np
    >>> mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_hermitian(mat)
    False

    References
    ==========
    .. [WikHerm] Wikipedia: Hermitian matrix.
        https://en.wikipedia.org/wiki/Hermitian_matrix

    :param mat: Matrix to check.
    :return: Return True if matrix is Hermitian, and False otherwise.
    """
    return np.allclose(mat, mat.conj().T)


def is_normal(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Determine if a matrix is normal [WikNormal]_.

    A matrix is normal if it commutes with its adjoint

    .. math::
        \begin{equation}
            [X, X^*] = 0,
        \end{equation}

    or, equivalently if

    .. math::
        \begin{equation}
            X^* X = X X^*.
        \end{equation}

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & 1
            \end{pmatrix}

    our function indicates that this is indeed a normal matrix.

    >>> from toqito.matrix_props import is_normal
    >>> import numpy as np
    >>> A = np.identity(4)
    >>> is_normal(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    is not normal.

    >>> from toqito.matrix_props import is_normal
    >>> import numpy as np
    >>> B = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_normal(B)
    False

    References
    ==========
    .. [WikNormal] Wikipedia: Normal matrix.
        https://en.wikipedia.org/wiki/Normal_matrix

    :param mat: The matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Returns True if the matrix is normal and False otherwise.
    """
    return np.allclose(
        np.matmul(mat, mat.conj().T), np.matmul(mat.conj().T, mat), rtol=rtol, atol=atol
    )


def is_pd(mat: np.ndarray) -> bool:
    r"""
    Check if matrix is positive definite (PD) [WikPD]_.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                2 & -1 & 0 \\
                -1 & 2 & -1 \\
                0 & -1 & 2
            \end{pmatrix}

    our function indicates that this is indeed a positive definite matrix.

    >>> from toqito.matrix_props import is_pd
    >>> import numpy as np
    >>> A = np.array([[2, -1, 0], [-1, 2, -1], [0, -1, 2]])
    >>> is_pd(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                -1 & -1 \\
                -1 & -1
            \end{pmatrix}

    is not positive definite.

    >>> from toqito.matrix_props import is_pd
    >>> import numpy as np
    >>> B = np.array([[-1, -1], [-1, -1]])
    >>> is_pd(B)
    False

    See Also
    ========
    is_psd

    References
    ==========
    .. [WikPD] Wikipedia: Definiteness of a matrix.
        https://en.wikipedia.org/wiki/Definiteness_of_a_matrix

    :param mat: Matrix to check.
    :return: Return True if matrix is PD, and False otherwise.
    """
    if np.array_equal(mat, mat.T):
        try:
            np.linalg.cholesky(mat)
            return True
        except np.linalg.LinAlgError:
            return False
    else:
        return False


def is_projection(mat: np.ndarray) -> bool:
    r"""
    Check if matrix is a projection matrix [WikProj]_.

    A matrix is a projection matrix if it is positive semidefinite (PSD) and if

    .. math::
        \begin{equation}
            X^2 = X
        \end{equation}

    where :math:`X` is the matrix in question.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                0 & 1 \\
                0 & 1
            \end{pmatrix}

    our function indicates that this is indeed a projection matrix.

    >>> from toqito.matrix_props import is_projection
    >>> import numpy as np
    >>> A = np.array([[0, 1], [0, 1]])
    >>> is_projection(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                -1 & -1 \\
                -1 & -1
            \end{pmatrix}

    is not positive definite.

    >>> from toqito.matrix_props import is_projection
    >>> import numpy as np
    >>> B = np.array([[-1, -1], [-1, -1]])
    >>> is_projection(B)
    False

    References
    ==========
    .. [WikProj] Wikipedia: Projection matrix.
        https://en.wikipedia.org/wiki/Projection_matrix

    :param mat: Matrix to check.
    :return: Return True if matrix is a projection matrix, and False otherwise.
    """
    if not is_psd(mat):
        return False
    return np.allclose(np.linalg.matrix_power(mat, 2), mat)


def is_psd(mat: np.ndarray, tol: float = 1e-8) -> bool:
    r"""
    Check if matrix is positive semidefinite (PSD) [WikPSD]_.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & -1 \\
                -1 & 1
            \end{pmatrix}

    our function indicates that this is indeed a positive semidefinite matrix.

    >>> from toqito.matrix_props import is_psd
    >>> import numpy as np
    >>> A = np.array([[1, -1], [-1, 1]])
    >>> is_psd(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                -1 & -1 \\
                -1 & -1
            \end{pmatrix}

    is not positive semidefinite.

    >>> from toqito.matrix_props import is_psd
    >>> import numpy as np
    >>> B = np.array([[-1, -1], [-1, -1]])
    >>> is_psd(B)
    False

    References
    ==========
    .. [WikPSD] Wikipedia: Definiteness of a matrix.
        https://en.wikipedia.org/wiki/Definiteness_of_a_matrix

    :param mat: Matrix to check.
    :param tol: Tolerance for numerical accuracy.
    :return: Return True if matrix is PSD, and False otherwise.
    """
    if not is_square(mat):
        return False
    return np.all(np.linalg.eigvalsh(mat) > -tol)


def is_square(mat: np.ndarray) -> bool:
    r"""
    Determine if a matrix is square [WikSquare]_.

    A matrix is square if the dimensions of the rows and columns are equivalent.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    our function indicates that this is indeed a square matrix.

    >>> from toqito.matrix_props import is_square
    >>> import numpy as np
    >>> A = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_square(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6
            \end{pmatrix}

    is not square.

    >>> from toqito.matrix_props import is_square
    >>> import numpy as np
    >>> B = np.array([[1, 2, 3], [4, 5, 6]])
    >>> is_square(B)
    False

    References
    ==========
    .. [WikSquare] Wikipedia: Square matrix.
        https://en.wikipedia.org/wiki/Square_matrix

    :param mat: The matrix to check.
    :return: Returns True if the matrix is square and False otherwise.
    """
    return mat.shape[0] == mat.shape[1]


def is_symmetric(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Determine if a matrix is symmetric [WikSym]_.

    The following 3x3 matrix is an example of a symmetric matrix:

    .. math::

        \begin{pmatrix}
            1 & 7 & 3 \\
            7 & 4 & -5 \\
            3 &-5 & 6
        \end{pmatrix}

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & 7 & 3 \\
                7 & 4 & -5 \\
                3 & -5 & 6
            \end{pmatrix}

    our function indicates that this is indeed a symmetric matrix.

    >>> from toqito.matrix_props import is_symmetric
    >>> import numpy as np
    >>> A = np.array([[1, 7, 3], [7, 4, -5], [3, -5, 6]])
    >>> is_symmetric(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 \\
                4 & 5
            \end{pmatrix}

    is not symmetric.

    >>> from toqito.matrix_props import is_symmetric
    >>> import numpy as np
    >>> B = np.array([[1, 2], [3, 4]])
    >>> is_symmetric(B)
    False

    References
    ==========
    .. [WikSym] Wikipedia: Symmetric matrix
        https://en.wikipedia.org/wiki/Symmetric_matrix

    :param mat: The matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Returns True if the matrix is symmetric and False otherwise.
    """
    return np.allclose(mat, mat.T, rtol=rtol, atol=atol)


def is_unitary(mat: Union[np.ndarray, np.matrix]) -> bool:
    r"""
    Check if matrix is unitary [WikUnitary]_.

    A matrix is unitary if its inverse is equal to its conjugate transpose.

    Alternatively, a complex square matrix :math:`U` is unitary if its conjugate
    transpose :math:`U^*` is also its inverse, that is, if

    .. math::
        \begin{equation}
            U^* U = U U^* = \mathbb{I},
        \end{equation}

    where :math:`\mathbb{I}` is the identity matrix.

    Examples
    ==========

    Consider the following matrix

    .. math::
        X = \begin{pmatrix}
            0 & 1 \\
            1 & 0
            \end{pmatrix}

    our function indicates that this is indeed a unitary matrix.

    >>> from toqito.matrix_props import is_unitary
    >>> import numpy as np
    >>> A = np.array([[0, 1], [1, 0]])
    >>> is_unitary(A)
    True

    We may also use the `random_unitary` function from `toqito`, and can verify
    that a randomly generated matrix is unitary

    >>> from toqito.matrix_props import is_unitary
    >>> from toqito.random import random_unitary
    >>> mat = random_unitary(2)
    >>> is_unitary(mat)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
            1 & 0 \\
            1 & 1
            \end{pmatrix}

    is not unitary.

    >>> from toqito.matrix_props import is_unitary
    >>> import numpy as np
    >>> B = np.array([[1, 0], [1, 1]])
    >>> is_unitary(B)
    False

    References
    ==========
    .. [WikUnitary] Wikipedia: Unitary matrix.
        https://en.wikipedia.org/wiki/Unitary_matrix

    :param mat: Matrix to check.
    :return: Return `True` if matrix is unitary, and `False` otherwise.
    """
    # If U^* * U = I U * U^*, the matrix "U" is unitary.
    return np.allclose(np.eye(len(mat)), mat.dot(mat.conj().T))



"""Tests for is_product_vector function."""
import unittest
import numpy as np

from toqito.states import basis
from toqito.states import max_entangled
from toqito.state_props import is_product_vector


class TestIsProductVector(unittest.TestCase):

    """Unit test for is_product_vector."""

    def test_is_product_entangled_state(self):
        """Check that is_product_vector returns False for an entangled state."""
        ent_vec = max_entangled(3)
        self.assertEqual(is_product_vector(ent_vec), False)

    def test_is_product_entangled_state_2_sys(self):
        """Check that dimension argument as list is supported."""
        ent_vec = max_entangled(4)
        self.assertEqual(is_product_vector(ent_vec, dim=[4, 4]), False)

    def test_is_product_entangled_state_3_sys(self):
        """Check that dimension argument as list is supported."""
        ent_vec = max_entangled(4)
        self.assertEqual(is_product_vector(ent_vec, dim=[2, 2, 2, 2]), False)

    def test_is_product_separable_state(self):
        """Check that is_product_vector returns True for a separable state."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        sep_vec = (
            1
            / 2
            * (
                np.kron(e_0, e_0)
                - np.kron(e_0, e_1)
                - np.kron(e_1, e_0)
                + np.kron(e_1, e_1)
            )
        )
        self.assertEqual(is_product_vector(sep_vec), True)


if __name__ == "__main__":
    unittest.main()



"""Two-player XOR game."""
import cvxpy
import numpy as np


class XORGame:
    r"""
    Create two-player XOR game object.

    Calculates the optimal probability that Alice and Bob win the game if they
    are allowed to determine a join strategy beforehand, but not allowed to
    communicate during the game itself.

    The quantum value of an XOR game can be solved via the semidefinite program
    from [CHTW04]_.

    This function is adapted from the QETLAB package.

    Examples
    ==========

    The CHSH game

    The CHSH game is a two-player nonlocal game with the following probability
    distribution and question and answer sets [CSUU08]_.

    .. math::
        \begin{equation}
            \begin{aligned} \pi(x,y) = \frac{1}{4}, \qquad (x,y) \in
                            \Sigma_A \times
                \Sigma_B, \qquad \text{and} \qquad (a, b) \in \Gamma_A \times
                \Gamma_B,
            \end{aligned}
        \end{equation}

    where

    .. math::
        \begin{equation}
            \Sigma_A = \{0, 1\}, \quad \Sigma_B = \{0, 1\}, \quad \Gamma_A =
            \{0,1\}, \quad \text{and} \quad \Gamma_B = \{0, 1\}.
        \end{equation}

    Alice and Bob win the CHSH game if and only if the following equation is
    satisfied

    .. math::
        \begin{equation}
        a \oplus b = x \land y.
        \end{equation}

    Recall that :math:`\oplus` refers to the XOR operation.

    The optimal quantum value of CHSH is :math:`\cos(\pi/8)^2 \approx 0.8536`
    where the optimal classical value is :math:`3/4`.

    In order to specify the CHSH game, we can define the probability matrix and
    predicate matrix for the CHSH game as `numpy` arrays as follows.

    >>> import numpy as np
    >>> prob_mat = np.array([[1 / 4, 1 / 4], [1 / 4, 1 / 4]])
    >>> pred_mat = np.array([[0, 0], [0, 1]])

    In `toqito`, we can calculate both the quantum and classical value of the
    CHSH game as follows.

    >>> import numpy as np
    >>> from toqito.nonlocal_games.xor_game import XORGame
    >>> chsh = XORGame(prob_mat, pred_mat)
    >>> chsh.quantum_value()
    0.8535533885683664
    >>>
    >>> chsh.classical_value()
    0.75

    The odd cycle game

    The odd cycle game is another XOR game [CHTW04]_. For this game, we can
    specify the probability and predicate matrices as follows.

    >>> prob_mat = np.array(
    >>> [
    >>>     [0.1, 0.1, 0, 0, 0],
    >>>     [0, 0.1, 0.1, 0, 0],
    >>>     [0, 0, 0.1, 0.1, 0],
    >>>     [0, 0, 0, 0.1, 0.1],
    >>>     [0.1, 0, 0, 0, 0.1],
    >>> ]
    >>> )
    >>> pred_mat = np.array(
    >>> [
    >>>     [0, 1, 0, 0, 0],
    >>>     [0, 0, 1, 0, 0],
    >>>     [0, 0, 0, 1, 0],
    >>>     [0, 0, 0, 0, 1],
    >>>     [1, 0, 0, 0, 0],
    >>> ]
    >>> )

    In `toqito`, we can calculate both the quantum and classical value of the
    odd cycle game as follows.

    >>> import numpy as np
    >>> from toqito.nonlocal_games.xor_game import XORGame
    >>> odd_cycle = XORGame(prob_mat, pred_mat)
    >>> odd_cycle.quantum_value()
    0.9755282544736033
    >>> odd_cycle.classical_value()
    0.9

    References
    ==========
    .. [CSUU08] Richard Cleve, William Slofstra, Falk Unger, Sarvagya Upadhyay
        "Strong parallel repetition theorem for quantum XOR proof systems",
        2008,
        https://arxiv.org/abs/quant-ph/0608146

    .. [CHTW04] Richard Cleve, Peter Hoyer, Ben Toner, John Watrous
        "Consequences and limits of nonlocal strategies."
        Proceedings. 19th IEEE Annual Conference on Computational Complexity,
        IEEE, 2004.
        https://arxiv.org/abs/quant-ph/0404076
    """

    def __init__(
        self, prob_mat: np.ndarray, pred_mat: np.ndarray, tol: float = None
    ) -> None:
        """
        Construct XOR game object.

        :param prob_mat: A matrix whose (q_0, q_1)-entry gives the probability that
                     the referee will give Alice the value `q_0` and Bob the
                     value `q_1`.
        :param pred_mat: A binary matrix whose (q_0, q_1)-entry indicates the
                     winning choice (either 0 or 1) when Alice and Bob receive
                     values `q_0` and `q_1` from the referee.
        :param tol: The error tolerance for the value.
        """
        self.prob_mat = prob_mat
        self.pred_mat = pred_mat

        q_0, q_1 = self.prob_mat.shape
        if tol is None:
            self.tol = np.finfo(float).eps * q_0 ** 2 * q_1 ** 2
        else:
            self.tol = tol

        # Perform some basic error checking to ensure the probability and
        # predicate matrices are well-defined.
        if (q_0, q_1) != self.pred_mat.shape:
            raise ValueError(
                "Invalid: The matrices `prob_mat` and `pred_mat` must"
                " be matrices of the same size."
            )
        if -np.min(np.min(self.prob_mat)) > self.tol:
            raise ValueError(
                "Invalid: The variable `prob_mat` must be a "
                "probability matrix: its entries must be "
                "non-negative."
            )
        if np.abs(np.sum(np.sum(self.prob_mat)) - 1) > self.tol:
            raise ValueError(
                "Invalid: The variable `prob_mat` must be a "
                "probability matrix: its entries must sum to 1."
            )

        self.quantum_strategy = None
        self.classical_strategy = None

    def quantum_value(self) -> float:
        """
        Compute the quantum value of the XOR game.

        :return: A value between [0, 1] representing the quantum value.
        """
        q_0, q_1 = self.prob_mat.shape

        # Compute the value of the game, depending on which type of value was
        # requested.
        # Use semidefinite programming to compute the value of the game.
        p_var = self.prob_mat * (1 - 2 * self.pred_mat)
        q_var = np.bmat(
            [[np.zeros((q_0, q_0)), p_var], [p_var.conj().T, np.zeros((q_1, q_1))]]
        )

        x_var = cvxpy.Variable((q_0 + q_1, q_0 + q_1), symmetric=True)
        objective = cvxpy.Maximize(cvxpy.trace(q_var @ x_var))
        constraints = [cvxpy.diag(x_var) == 1, x_var >> 0]
        problem = cvxpy.Problem(objective, constraints)

        problem.solve()

        return np.real(problem.value) / 4 + 1 / 2

    def classical_value(self) -> float:
        """
        Compute the classical value of the XOR game.

        :return: A value between [0, 1] representing the classical value.
        """
        q_0, q_1 = self.prob_mat.shape

        # At worst, out winning probability is 0. Now, try to improve.
        val = 0

        # Find the maximum probability of winning (this is NP-hard, so don't
        # expect an easy way to do it: just loop over all strategies.

        # Loop over Alice's answers
        for a_ans in range(2 ** q_0):
            # Loop over Bob's answers:
            for b_ans in range(2 ** q_1):
                a_vec = (a_ans >> np.arange(q_0)) & 1
                b_vec = (b_ans >> np.arange(q_1)) & 1

                # Now compute the winning probability under this strategy: XOR
                # together Alice's responses and Bob's responses, then check
                # where the XORed value equals the value in the given matrix F.
                # Where the values match, multiply by the probability of
                # getting that pair of questions (i.e., multiply by the
                # probability of getting that pair of questions (i.e., multiply
                # entry-wise by P) and then sum over the rows and columns.
                self.classical_strategy = np.mod(
                    np.multiply(a_vec.conj().T.reshape(-1, 1), np.ones((1, q_1)))
                    + np.multiply(np.ones((q_0, 1)), b_vec),
                    2,
                )
                p_win = np.sum(
                    np.sum(
                        np.multiply(
                            self.classical_strategy == self.pred_mat, self.prob_mat
                        )
                    )
                )

                # Is this strategy better than other ones tried so far?
                val = max(val, p_win)

                # Already optimal? Quit.
                if val >= 1 - self.tol:
                    return val
        return val



"""Tests for gen_gell_man function."""
import unittest
import numpy as np

from toqito.matrices import gen_gell_mann


class TestGenGellMann(unittest.TestCase):

    """Unit test for gen_gell_mann."""

    def test_gell_mann_identity(self):
        """Generalized Gell-Mann operator identity."""
        expected_res = np.array([[1, 0], [0, 1]])
        res = gen_gell_mann(0, 0, 2)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_pauli_x(self):
        """Generalized Gell-Mann operator Pauli-X."""
        expected_res = np.array([[0, 1], [1, 0]])
        res = gen_gell_mann(0, 1, 2)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_pauli_y(self):
        """Generalized Gell-Mann operator Pauli-Y."""
        expected_res = np.array([[0, -1j], [1j, 0]])
        res = gen_gell_mann(1, 0, 2)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_pauli_z(self):
        """Generalized Gell-Mann operator Pauli-Z."""
        expected_res = np.array([[1, 0], [0, -1]])
        res = gen_gell_mann(1, 1, 2)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_dim_3_1(self):
        """Generalized Gell-Mann operator 3-dimensional."""
        expected_res = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
        res = gen_gell_mann(0, 1, 3)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_dim_3_2(self):
        """Generalized Gell-Mann operator 3-dimensional."""
        expected_res = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])
        res = gen_gell_mann(0, 2, 3)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_dim_3_3(self):
        """Generalized Gell-Mann operator 3-dimensional."""
        expected_res = np.array(
            [
                [1 / np.sqrt(3), 0, 0],
                [0, 1 / np.sqrt(3), 0],
                [0, 0, -2 * 1 / np.sqrt(3)],
            ]
        )
        res = gen_gell_mann(2, 2, 3)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_dim_4_1(self):
        """Generalized Gell-Mann operator 4-dimensional."""
        expected_res = np.array(
            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]
        )
        res = gen_gell_mann(2, 3, 4)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_gell_mann_sparse(self):
        """Generalized Gell-Mann operator sparse."""
        res = gen_gell_mann(205, 34, 500, True)

        self.assertEqual(res[34, 205], -1j)
        self.assertEqual(res[205, 34], 1j)


if __name__ == "__main__":
    unittest.main()



"""Tests for random_unitary function."""
import unittest

from toqito.random import random_unitary
from toqito.matrix_props import is_unitary


class TestRandomUnitary(unittest.TestCase):

    """Unit test for random_unitary."""

    def test_random_unitary_not_real(self):
        """Generate random non-real unitary matrix."""
        mat = random_unitary(2)
        self.assertEqual(is_unitary(mat), True)

    def test_random_unitary_real(self):
        """Generate random real unitary matrix."""
        mat = random_unitary(2, True)
        self.assertEqual(is_unitary(mat), True)

    def test_random_unitary_vec_dim(self):
        """Generate random non-real unitary matrix."""
        mat = random_unitary([4, 4], True)
        self.assertEqual(is_unitary(mat), True)


if __name__ == "__main__":
    unittest.main()



"""Operations on quantum states."""
from typing import List, Tuple, Union
from scipy.sparse import issparse, linalg

import numpy as np


__all__ = ["pure_to_mixed",
           "schmidt_decomposition",
           "tensor"]


def pure_to_mixed(phi: np.ndarray) -> np.ndarray:
    r"""
    Convert a state vector or density matrix to a density matrix.

    Examples
    ==========

    It is possible to convert a pure state vector to a mixed state vector using
    the `toqito` package. Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right).

    The corresponding mixed state from :math:`u` is calculated as

    .. math::
        \rho = u u^* = \frac{1}{2} \begin{pmatrix}
                                        1 & 0 & 0 & 1 \\
                                        0 & 0 & 0 & 0 \\
                                        0 & 0 & 0 & 0 \\
                                        1 & 0 & 0 & 1
                                   \end{pmatrix}

    Using `toqito`, we can obtain this matrix as follows.

    >>> from toqito.states import bell
    >>> from toqito.state_ops import pure_to_mixed
    >>> phi = bell(0)
    >>> pure_to_mixed(phi)
    [[0.5, 0. , 0. , 0.5],
     [0. , 0. , 0. , 0. ],
     [0. , 0. , 0. , 0. ],
     [0.5, 0. , 0. , 0.5]]

    We can also give matrix inputs to the function in `toqito`.

    >>> from toqito.states import bell
    >>> from toqito.state_ops import pure_to_mixed
    >>> phi = bell(0) * bell(0).conj().T
    >>> pure_to_mixed(phi)
    [[0.5, 0. , 0. , 0.5],
     [0. , 0. , 0. , 0. ],
     [0. , 0. , 0. , 0. ],
     [0.5, 0. , 0. , 0.5]])

    :param phi: A density matrix or a pure state vector.
    :return: density matrix representation of `phi`, regardless of
             whether `phi` is itself already a density matrix or if
             if is a pure state vector.
    """
    # Compute the size of `phi`. If it's already a mixed state, leave it alone.
    # If it's a vector (pure state), make it into a density matrix.
    row_dim, col_dim = phi.shape[0], phi.shape[1]

    # It's a pure state vector.
    if min(row_dim, col_dim) == 1:
        return phi * phi.conj().T
    # It's a density matrix.
    if row_dim == col_dim:
        return phi
    # It's neither.
    raise ValueError("InvalidDim: `phi` must be either a vector or square " "matrix.")


def schmidt_decomposition(
    vec: np.ndarray, dim: Union[int, List[int], np.ndarray] = None, k_param: int = 0
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""
    Compute the Schmidt decomposition of a bipartite vector [WikSD]_.

    Examples
    ==========

    Consider the :math:`3`-dimensional maximally entangled state

    .. math::
        u = \frac{1}{\sqrt{3}} \left( |000 \rangle +
        |111 \rangle + |222 \rangle \right)

    We can generate this state using the `toqito` module as follows.

    >>> from toqito.states import max_entangled
    >>> max_entangled(3)
    [[0.57735027],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.57735027],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.57735027]]

    Computing the Schmidt decomposition of :math:`u`, we can obtain the
    corresponding singular values of :math:`u` as

    .. math::
        \frac{1}{\sqrt{3}} \left[1, 1, 1 \right]^{\text{T}}

    >>> from toqito.states import max_entangled
    >>> from toqito.state_ops import schmidt_decomposition
    >>> singular_vals, u_mat, vt_mat = schmidt_decomposition(max_entangled(3))
    >>> singular_vals
    [[0.57735027]
     [0.57735027]
     [0.57735027]]
    >>> u_mat
    [[1. 0. 0.]
     [0. 1. 0.]
     [0. 0. 1.]]
    >>> vt_mat
    [[1. 0. 0.]
     [0. 1. 0.]
     [0. 0. 1.]]

    References
    ==========
    .. [WikSD] Wikipedia: Schmidt decomposition
        https://en.wikipedia.org/wiki/Schmidt_decomposition

    :param vec:
    :param dim:
    :param k_param:
    :return: The Schmidt decomposition of the `vec` input.
    """
    eps = np.finfo(float).eps

    if dim is None:
        dim = np.round(np.sqrt(len(vec)))
    if dim is list:
        dim = np.array(dim)

    # Allow the user to enter a single number for `dim`.
    if isinstance(dim, float):
        dim = np.array([dim, len(vec) / dim])
        if np.abs(dim[1] - np.round(dim[1])) >= 2 * len(vec) * eps:
            raise ValueError(
                "InvalidDim: The value of `dim` must evenly divide"
                " `len(vec)`; please provide a `dim` array "
                "containing the dimensions of the subsystems."
            )
        dim[1] = np.round(dim[1])

    # Try to guess whether SVD or SVDS will be faster, and then perform the
    # appropriate singular value decomposition.
    adj = 20 + 1000 * (not issparse(vec))

    # Just a few Schmidt coefficients.
    if 0 < k_param <= np.ceil(np.min(dim) / adj):
        u_mat, singular_vals, vt_mat = linalg.svds(
            linalg.LinearOperator(np.reshape(vec, dim[::-1].astype(int)), k_param)
        )
    # Otherwise, use lots of Schmidt coefficients.
    else:
        u_mat, singular_vals, vt_mat = np.linalg.svd(
            np.reshape(vec, dim[::-1].astype(int))
        )

    if k_param > 0:
        u_mat = u_mat[:, :k_param]
        singular_vals = singular_vals[:k_param]
        vt_mat = vt_mat[:, :k_param]

    # singular_vals = np.diag(singular_vals)
    singular_vals = singular_vals.reshape(-1, 1)
    if k_param == 0:
        # Schmidt rank.
        r_param = np.sum(singular_vals > np.max(dim) * np.spacing(singular_vals[0]))
        # Schmidt coefficients.
        singular_vals = singular_vals[:r_param]
        u_mat = u_mat[:, :r_param]
        vt_mat = vt_mat[:, :r_param]

    u_mat = u_mat.conj().T
    return singular_vals, u_mat, vt_mat


def tensor(*args) -> np.ndarray:
    r"""
    Compute the Kronecker tensor product [WikTensor]_.

    Tensor two matrices or vectors together using the standard Kronecker
    operation provided from numpy.

    Given two matrices :math:`A` and :math:`B`, computes :math:`A \otimes B`.
    The same concept also applies to two vectors :math:`v` and :math:`w` which
    computes :math:`v \otimes w`.

    One may also compute the tensor product one matrix `n` times with itself.

    For a matrix, :math:`A` and an integer :math:`n`, the result of this
    function computes :math:`A^{\otimes n}`.

    Similarly for a vector :math:`v` and an integer :math:`n`, the result of
    of this function computes :math:`v^{\otimes n}`.

    One may also perform the tensor product on a list of matrices.

    Given a list of :math:`n` matrices :math:`A_1, A_2, \ldots, A_n` the result
    of this function computes :math:`A_1 \otimes A_2 \otimes \ldots
    \otimes A_n`.

    Similarly, for a list of :math:`n` vectors :math:`v_1, v_2, \ldots, v_n`,
    the result of this function computes :math:`v_1 \otimes v_2 \otimes \ldots
    \otimes v_n`.

    Examples
    ==========

    Tensor product two matrices or vectors

    Consider the following ket vector

    .. math::
        |0 \rangle = \left[1, 0 \right]^{\text{T}}

    Computing the following tensor product

    .. math:
        |0 \rangle \otimes |0 \rangle = \left[1, 0, 0, 0 \right]^{\text{T}}

    This can be accomplished in `toqito` as follows.

    >>> from toqito.states import basis
    >>> from toqito.state_ops import tensor
    >>> e_0 = basis(2, 0)
    >>> tensor(e_0, e_0)
    [[1],
     [0],
     [0],
     [0]]

    Tensor product one matrix :math:`n` times with itself.

    We may also tensor some element with itself some integer number of times.
    For instance we can compute

    .. math::
        |0 \rangle^{\otimes 3} = \left[1, 0, 0, 0, 0, 0, 0, 0 \right]^{\text{T}}

    in `toqito` as follows.

    >>> from toqito.states import basis
    >>> from toqito.state_ops.tensor import tensor
    >>> e_0 = basis(2, 0)
    >>> tensor(e_0, 3)
    [[1],
     [0],
     [0],
     [0],
     [0],
     [0],
     [0],
     [0]]

    Perform the tensor product on a list of vectors or matrices.

    If we wish to compute the tensor product against more than two matrices or
    vectors, we can feed them in as a `list`. For instance, if we wish to
    compute :math:`|0 \rangle \otimes |1 \rangle \otimes |0 \rangle`, we can do
    so as follows.

    >>> from toqito.states import basis
    >>> from toqito.state_ops.tensor import tensor
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> tensor([e_0, e_1, e_0])
    [[0],
     [0],
     [1],
     [0],
     [0],
     [0],
     [0],
     [0]]

    References
    ==========
    .. [WikTensor] Wikipedia: Tensor product
        https://en.wikipedia.org/wiki/Tensor_product

    :param args: Input to the tensor function is expected to be either:
        - List[np.ndarray]: List of numpy matrices,
        - np.ndarray, ... np.ndarray: An arbitrary number of numpy arrays,
        - np.ndarray, int: A numpy array and an integer.
    :return: The computed tensor product.
    """
    result = None

    # Input is provided as a list of matrices.
    if len(args) == 1 and isinstance(args[0], list):
        if len(args[0]) == 1:
            return args[0][0]
        if len(args[0]) == 2:
            return np.kron(args[0][0], args[0][1])
        if len(args[0]) >= 3:
            result = args[0][0]
            for i in range(1, len(args[0])):
                result = np.kron(result, args[0][i])
        return result

    # Tensor product one matrix `n` times with itself.
    if len(args) == 2 and isinstance(args[1], int):
        num_tensor = args[1]
        if num_tensor == 1:
            return args[0]
        if num_tensor == 2:
            return np.kron(args[0], args[0])
        if num_tensor >= 3:
            result = np.kron(args[0], args[0])
            for _ in range(2, num_tensor):
                result = np.kron(result, args[0])
        return result

    # Tensor product between two or more matrices.
    if len(args) == 1:
        return args[0]
    if len(args) == 2:
        return np.kron(args[0], args[1])
    if len(args) >= 3:
        result = args[0]
        for i in range(1, len(args)):
            result = np.kron(result, args[i])
    return result



"""Tests for ghz function."""
import unittest
import numpy as np

from toqito.states import basis, ghz
from toqito.state_ops import tensor


class TestGHZ(unittest.TestCase):

    """Unit test for ghz."""

    def test_ghz_2_3(self):
        """Produces the 3-qubit GHZ state: `1/sqrt(2) * (|000> + |111>)`."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        expected_res = 1 / np.sqrt(2) * (tensor(e_0, e_0, e_0) + tensor(e_1, e_1, e_1))

        res = ghz(2, 3).toarray()

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_ghz_4_7(self):
        r"""
        The following generates the following GHZ state in `(C^4)^{\otimes 7}`.

        `1/sqrt(30) * (|0000000> + 2|1111111> + 3|2222222> + 4|3333333>)`.
        """
        e0_4 = np.array([[1], [0], [0], [0]])
        e1_4 = np.array([[0], [1], [0], [0]])
        e2_4 = np.array([[0], [0], [1], [0]])
        e3_4 = np.array([[0], [0], [0], [1]])

        expected_res = (
            1
            / np.sqrt(30)
            * (
                tensor(e0_4, e0_4, e0_4, e0_4, e0_4, e0_4, e0_4)
                + 2 * tensor(e1_4, e1_4, e1_4, e1_4, e1_4, e1_4, e1_4)
                + 3 * tensor(e2_4, e2_4, e2_4, e2_4, e2_4, e2_4, e2_4)
                + 4 * tensor(e3_4, e3_4, e3_4, e3_4, e3_4, e3_4, e3_4)
            )
        )

        res = ghz(4, 7, np.array([1, 2, 3, 4]) / np.sqrt(30)).toarray()

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_invalid_dim(self):
        """Tests for invalid dimensions."""
        with self.assertRaises(ValueError):
            ghz(1, 2)

    def test_invalid_qubits(self):
        """Tests for invalid number of qubits."""
        with self.assertRaises(ValueError):
            ghz(2, 1)

    def test_invalid_coeff(self):
        """Tests for invalid coefficients."""
        with self.assertRaises(ValueError):
            ghz(2, 3, [1, 2, 3, 4, 5])


if __name__ == "__main__":
    unittest.main()



"""Tests for negativity function."""
import unittest
import numpy as np

from toqito.state_props import negativity


class TestNegativity(unittest.TestCase):

    """Unit test for negativity."""

    def test_negativity_rho(self):
        """Test for negativity on rho."""
        test_input_mat = np.array(
            [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
        )

        self.assertEqual(np.isclose(negativity(test_input_mat), 1 / 2), True)

    def test_negativity_rho_dim_int(self):
        """Test for negativity on rho."""
        test_input_mat = np.array(
            [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
        )

        self.assertEqual(np.isclose(negativity(test_input_mat, 2), 1 / 2), True)

    def test_invalid_rho_dim_int(self):
        """Invalid dim parameters."""
        with self.assertRaises(ValueError):
            test_input_mat = np.array(
                [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
            )
            negativity(test_input_mat, 5)

    def test_invalid_rho_dim_vec(self):
        """Invalid dim parameters."""
        with self.assertRaises(ValueError):
            test_input_mat = np.array(
                [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
            )
            negativity(test_input_mat, [2, 5])


if __name__ == "__main__":
    unittest.main()



"""Tests for is_ppt function."""
import unittest
import numpy as np

from toqito.states import bell
from toqito.state_props import is_ppt


class TestIsPPT(unittest.TestCase):

    """Unit test for is_ppt."""

    def test_is_ppt(self):
        """Check that PPT matrix returns True."""
        mat = np.identity(9)
        self.assertEqual(is_ppt(mat), True)

    def test_is_ppt_sys(self):
        """Check that PPT matrix returns True with sys specified."""
        mat = np.identity(9)
        self.assertEqual(is_ppt(mat, 2), True)

    def test_is_ppt_dim_sys(self):
        """Check that PPT matrix returns True with dim and sys specified."""
        mat = np.identity(9)
        self.assertEqual(is_ppt(mat, 2, np.round(np.sqrt(mat.size))), True)

    def test_is_ppt_tol(self):
        """Check that PPT matrix returns True."""
        mat = np.identity(9)
        self.assertEqual(is_ppt(mat, 2, np.round(np.sqrt(mat.size)), 1e-10), True)

    def test_entangled_state(self):
        """Entangled state of dimension 2 will violate PPT criterion."""
        rho = bell(2) * bell(2).conj().T
        self.assertEqual(is_ppt(rho), False)


if __name__ == "__main__":
    unittest.main()



"""Tests for vec function."""
import unittest
import numpy as np

from toqito.matrix_ops import vec


class TestVec(unittest.TestCase):

    """Unit test for vec."""

    def test_vec(self):
        """Test standard vec operation on a matrix."""
        expected_res = np.array([[1], [3], [2], [4]])

        test_input_mat = np.array([[1, 2], [3, 4]])

        res = vec(test_input_mat)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for von_neumann_entropy function."""
import unittest
import numpy as np

from toqito.states import bell
from toqito.states import max_mixed
from toqito.state_metrics import von_neumann_entropy


class TestVonNeumannEntropy(unittest.TestCase):

    """Unit tests for von_neumann_entropy."""

    def test_von_neumann_entropy_bell_state(self):
        """Entangled state von Neumann entropy should be zero."""
        rho = bell(0) * bell(0).conj().T
        res = von_neumann_entropy(rho)
        self.assertEqual(np.isclose(res, 0), True)

    def test_von_neumann_entropy_max_mixed_statre(self):
        """Von Neumann entropy of the maximally mixed state should be one."""
        res = von_neumann_entropy(max_mixed(2, is_sparse=False))
        self.assertEqual(np.isclose(res, 1), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for partial_transpose function."""
import unittest
import numpy as np

from toqito.states import bell
from toqito.channels import partial_transpose


class TestPartialTranspose(unittest.TestCase):

    """Unit test for partial_transpose."""

    def test_partial_transpose(self):
        """
        Default partial_transpose.

        By default, the partial_transpose function performs the transposition
        on the second subsystem.
        """
        test_input_mat = np.arange(1, 17).reshape(4, 4)

        expected_res = np.array(
            [[1, 5, 3, 7], [2, 6, 4, 8], [9, 13, 11, 15], [10, 14, 12, 16]]
        )

        res = partial_transpose(test_input_mat)

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_partial_transpose_sys(self):
        """
        Default partial transpose `sys` argument.

        By specifying the `sys` argument, you can perform the transposition on
        the first subsystem instead:
        """
        test_input_mat = np.arange(1, 17).reshape(4, 4)

        expected_res = np.array(
            [[1, 2, 9, 10], [5, 6, 13, 14], [3, 4, 11, 12], [7, 8, 15, 16]]
        )

        res = partial_transpose(test_input_mat, 1)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_partial_transpose_sys_vec(self):
        """Partial transpose on matrix with `sys` defined as vector."""
        test_input_mat = np.arange(1, 17).reshape(4, 4)

        expected_res = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )

        res = partial_transpose(test_input_mat, [1, 2])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_partial_transpose_sys_vec_dim_vec(self):
        """Variables `sys` and `dim` defined as vector."""
        test_input_mat = np.arange(1, 17).reshape(4, 4)

        expected_res = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )

        res = partial_transpose(test_input_mat, [1, 2], [2, 2])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_partial_transpose_norm_diff(self):
        """
        Apply partial transpose to first and second subsystem.

        Applying the transpose to both the first and second subsystems results
        in the standard transpose of the matrix.
        """
        test_input_mat = np.arange(1, 17).reshape(4, 4)
        res = np.linalg.norm(
            partial_transpose(test_input_mat, [1, 2]) - test_input_mat.conj().T
        )
        expected_res = 0

        self.assertEqual(np.isclose(res, expected_res), True)

    def test_partial_transpose_16_by_16(self):
        """Partial transpose on a 16-by-16 matrix."""
        test_input_mat = np.arange(1, 257).reshape(16, 16)
        res = partial_transpose(test_input_mat, [1, 3], [2, 2, 2, 2])
        first_expected_row = np.array(
            [1, 2, 33, 34, 5, 6, 37, 38, 129, 130, 161, 162, 133, 134, 165, 166]
        )

        first_expected_col = np.array(
            [1, 17, 3, 19, 65, 81, 67, 83, 9, 25, 11, 27, 73, 89, 75, 91]
        )

        self.assertEqual(np.allclose(res[0, :], first_expected_row), True)
        self.assertEqual(np.allclose(res[:, 0], first_expected_col), True)

    def test_bell_state_pt(self):
        """Test partial transpose on a Bell state."""
        rho = bell(2) * bell(2).conj().T
        expected_res = np.array(
            [[0, 0, 0, 1 / 2], [0, 1 / 2, 0, 0], [0, 0, 1 / 2, 0], [1 / 2, 0, 0, 0]]
        )
        res = partial_transpose(rho)
        self.assertEqual(np.allclose(res, expected_res), True)

    def test_non_square_matrix(self):
        """Matrix must be square."""
        with self.assertRaises(ValueError):
            test_input_mat = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [13, 14, 15, 16]])
            partial_transpose(test_input_mat)

    def test_non_square_matrix_2(self):
        """Matrix must be square."""
        with self.assertRaises(ValueError):
            rho = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
            partial_transpose(rho, 2, [2])


if __name__ == "__main__":
    unittest.main()



"""Tests for antisymmetric_projection function."""
import unittest
import numpy as np

from toqito.perms import antisymmetric_projection


class TestAntisymmetricProjection(unittest.TestCase):

    """Unit test for antisymmetric_projection."""

    def test_antisymmetric_projection_d_2_p_1(self):
        """Dimension is 2 and p is equal to 1."""
        res = antisymmetric_projection(2, 1).todense()
        expected_res = np.array([[1, 0], [0, 1]])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_antisymmetric_projection_p_larger_than_d(self):
        """The `p` value is greater than the dimension `d`."""
        res = antisymmetric_projection(2, 3).todense()
        expected_res = np.zeros((8, 8))

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_antisymmetric_projection_2(self):
        """The dimension is 2."""
        res = antisymmetric_projection(2).todense()
        expected_res = np.array(
            [[0, 0, 0, 0], [0, 0.5, -0.5, 0], [0, -0.5, 0.5, 0], [0, 0, 0, 0]]
        )

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_antisymmetric_projection_3_3_true(self):
        """The `dim` is 3, the `p` is 3, and `partial` is True."""
        res = antisymmetric_projection(3, 3, True).todense()
        self.assertEqual(np.isclose(res[5].item(), -0.40824829), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for is_mub function."""
import unittest
import numpy as np

from toqito.states import basis
from toqito.state_props import is_mub


class TestIsMUB(unittest.TestCase):

    """Unit test for is_mub."""

    def test_is_mub_dim_2(self):
        """Return True for MUB of dimension 2."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        mub_1 = [e_0, e_1]
        mub_2 = [1 / np.sqrt(2) * (e_0 + e_1), 1 / np.sqrt(2) * (e_0 - e_1)]
        mub_3 = [1 / np.sqrt(2) * (e_0 + 1j * e_1), 1 / np.sqrt(2) * (e_0 - 1j * e_1)]
        mubs = [mub_1, mub_2, mub_3]
        self.assertEqual(is_mub(mubs), True)

    def test_is_not_mub_dim_2(self):
        """Return False for non-MUB of dimension 2."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        mub_1 = [e_0, e_1]
        mub_2 = [1 / np.sqrt(2) * (e_0 + e_1), e_1]
        mub_3 = [1 / np.sqrt(2) * (e_0 + 1j * e_1), e_0]
        mubs = [mub_1, mub_2, mub_3]
        self.assertEqual(is_mub(mubs), False)

    def test_invalid_input_len(self):
        """Tests for invalid input len."""
        with self.assertRaises(ValueError):
            vec_list = [np.array([1, 0])]
            is_mub(vec_list)


if __name__ == "__main__":
    unittest.main()



"""Tests for swap function."""
import unittest
import numpy as np

from toqito.perms import swap


class TestSwap(unittest.TestCase):

    """Unit test for swap."""

    def test_swap_matrix(self):
        """Tests swap operation on matrix."""
        test_mat = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )

        expected_res = np.array(
            [[1, 9, 5, 13], [3, 11, 7, 15], [2, 10, 6, 14], [4, 12, 8, 16]]
        )

        res = swap(test_mat)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_swap_vector_1(self):
        """Tests swap operation on vector."""
        test_vec = np.array([1, 2, 3, 4])

        expected_res = np.array([1, 3, 2, 4])

        res = swap(test_vec)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_swap_int_dim(self):
        """Test swap operation when int is provided."""
        test_mat = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )

        expected_res = np.array(
            [[1, 9, 5, 13], [3, 11, 7, 15], [2, 10, 6, 14], [4, 12, 8, 16]]
        )

        res = swap(test_mat, [1, 2], 2)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_invalid_dim(self):
        """Invalid dim parameters."""
        test_mat = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )
        with self.assertRaises(ValueError):
            swap(test_mat, [1, 2], 5)

    def test_invalid_sys(self):
        """Invalid sys parameters."""
        test_mat = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )
        with self.assertRaises(ValueError):
            swap(test_mat, [0])

    def test_invalid_sys_len(self):
        """Invalid sys parameters."""
        test_mat = np.array(
            [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
        )
        with self.assertRaises(ValueError):
            swap(test_mat, [1], 2)


if __name__ == "__main__":
    unittest.main()



"""Tests for hilbert_schmidt function."""
import unittest
import numpy as np
from toqito.states import bell
from toqito.state_metrics import hilbert_schmidt


class TestHilbertSchmidt(unittest.TestCase):

    """Unit test for hilbert_schmidt."""

    def test_hilbert_schmidt_bell(self):
        r"""Test Hilbert-Schmidt distance on two Bell states."""

        rho = bell(0) * bell(0).conj().T
        sigma = bell(3) * bell(3).conj().T

        res = hilbert_schmidt(rho, sigma)

        self.assertEqual(np.isclose(res, 1), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for reduction_map function."""
import unittest

from toqito.channels import reduction


class TestReduction(unittest.TestCase):

    """Unit test for reduction."""

    def test_reduction_map(self):
        """Test for the standard reduction map."""
        res = reduction(3)
        self.assertEqual(res[4, 0], -1)
        self.assertEqual(res[8, 0], -1)
        self.assertEqual(res[1, 1], 1)
        self.assertEqual(res[2, 2], 1)
        self.assertEqual(res[3, 3], 1)
        self.assertEqual(res[0, 4], -1)
        self.assertEqual(res[8, 4], -1)
        self.assertEqual(res[5, 5], 1)
        self.assertEqual(res[6, 6], 1)
        self.assertEqual(res[7, 7], 1)
        self.assertEqual(res[0, 8], -1)
        self.assertEqual(res[4, 8], -1)


if __name__ == "__main__":
    unittest.main()



"""Tests for is_commuting function."""
import unittest
import numpy as np

from toqito.matrix_props import is_commuting


class TestIsCommuting(unittest.TestCase):

    """Unit test for is_commuting."""

    def test_is_commuting_false(self):
        """Test if non-commuting matrices return False."""
        mat_1 = np.array([[0, 1], [0, 0]])
        mat_2 = np.array([[1, 0], [0, 0]])
        self.assertEqual(is_commuting(mat_1, mat_2), False)

    def test_is_commuting_true(self):
        """Test commuting matrices return True."""
        mat_1 = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 2]])
        mat_2 = np.array([[2, 4, 0], [3, 1, 0], [-1, -4, 1]])
        self.assertEqual(is_commuting(mat_1, mat_2), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for is_herm_preserving function."""
import unittest
import numpy as np

from toqito.channel_props import is_herm_preserving
from toqito.perms import swap_operator


class TestIsHermPreserving(unittest.TestCase):

    """Unit test for is_herm_preserving."""

    def test_is_herm_preserving_kraus_true(self):
        """Verify Hermitian-preserving channel as Kraus ops as True."""
        unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
        kraus_ops = [[np.identity(2), np.identity(2)], [unitary_mat, -unitary_mat]]

        self.assertEqual(is_herm_preserving(kraus_ops), True)

    def test_is_herm_preserving_kraus_false(self):
        """Verify non-Hermitian-preserving channel as Kraus ops as False."""
        unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
        kraus_ops = [[np.identity(2), unitary_mat], [unitary_mat, -unitary_mat]]

        self.assertEqual(is_herm_preserving(kraus_ops), False)

    def test_is_herm_preserving_choi_true(self):
        """
        The swap operator is the Choi matrix of the (Hermitian-preserving)
        transpose map.
        """
        choi_mat = swap_operator(3)
        self.assertEqual(is_herm_preserving(choi_mat), True)

    def test_is_herm_preserving_non_square(self):
        """Verify non-square Choi matrix returns False."""
        non_square_mat = np.array([[1, 2, 3], [4, 5, 6]])
        self.assertEqual(is_herm_preserving(non_square_mat), False)


if __name__ == "__main__":
    unittest.main()



"""Tests for NonlocalGame class."""
import unittest
import numpy as np

from toqito.nonlocal_games.nonlocal_game import NonlocalGame


class TestNonlocalGame(unittest.TestCase):

    """Unit test for NonlocalGame."""

    def test_chsh_lower_bound(self):
        """Calculate the lower bound on the quantum value for the CHSH game."""
        dim = 2
        num_alice_inputs, num_alice_outputs = 2, 2
        num_bob_inputs, num_bob_outputs = 2, 2
        prob_mat = np.array([[1 / 4, 1 / 4], [1 / 4, 1 / 4]])
        pred_mat = np.zeros(
            (num_alice_outputs, num_bob_outputs, num_alice_inputs, num_bob_inputs)
        )

        for a_alice in range(num_alice_outputs):
            for b_bob in range(num_bob_outputs):
                for x_alice in range(num_alice_inputs):
                    for y_bob in range(num_bob_inputs):
                        if np.mod(a_alice + b_bob + x_alice * y_bob, dim) == 0:
                            pred_mat[a_alice, b_bob, x_alice, y_bob] = 1

        chsh = NonlocalGame(dim, prob_mat, pred_mat)
        res = chsh.quantum_value_lower_bound()
        self.assertEqual(np.isclose(res, np.cos(np.pi / 8) ** 2, rtol=1e-02), True)

    def test_ffl_game_classical_value(self):
        """Classical value for the FFL game."""
        dim = 2
        num_alice_in, num_alice_out = 2, 2
        num_bob_in, num_bob_out = 2, 2
        prob_mat = np.array([[1 / 3, 1 / 3], [1 / 3, 0]])

        pred_mat = np.zeros((num_alice_out, num_bob_out, num_alice_in, num_bob_in))
        for a_alice in range(num_alice_out):
            for b_bob in range(num_bob_out):
                for x_alice in range(num_alice_in):
                    for y_bob in range(num_bob_in):
                        if (a_alice or x_alice) != (b_bob or y_bob):
                            pred_mat[a_alice, b_bob, x_alice, y_bob] = 1

        ffl = NonlocalGame(dim, prob_mat, pred_mat)
        res = ffl.classical_value()
        expected_res = 2 / 3
        self.assertEqual(np.isclose(res, expected_res), True)


if __name__ == "__main__":
    unittest.main()



"""Properties of measurement operators."""
from typing import List
import numpy as np

from toqito.matrix_props import is_psd

__all__ = ["is_povm"]


def is_povm(mat_list: List[np.ndarray]) -> bool:
    r"""
    Determine if a list of matrices constitute a valid set of POVMs [WikPOVM]_.

    A valid set of measurements are defined by a set of positive semidefinite
    operators

    .. math::
         \{P_a : a \in \Gamma\} \subset \text{Pos}(\mathcal{X})

    indexed by the alphabet :math:`\Gamma` of measurement outcomes satisfying
    the constraint that

    .. math::
        \sum_{a \in \Gamma} P_a = I_{\mathcal{X}}.

    Examples
    ==========

    Consider the following matrices

    .. math::
        M_0 = \begin{pmatrix}
                                1 & 0 \\
                                0 & 0
                             \end{pmatrix}, \qquad
        M_1 = \begin{pmatrix}
                0 & 0 \\
                0 & 1
              \end{pmatrix}

    our function indicates that this set of operators constitute a POVM.

    >>> from toqito.measurement_props import is_povm
    >>> import numpy as np
    >>> meas_1 = np.array([[1, 0], [0, 0]])
    >>> meas_2 = np.array([[0, 0], [0, 1]])
    >>> meas = [meas_1, meas_2]
    >>> is_povm(meas)
    True

    We may also use the `random_povm` function from `toqito`, and can verify
    that a randomly generated set satisfies the criteria for being a POVM set.

    >>> from toqito.measurement_props import is_povm
    >>> from toqito.random import random_povm
    >>> import numpy as np
    >>> dim, num_inputs, num_outputs = 2, 2, 2
    >>> measurements = random_povm(dim, num_inputs, num_outputs)
    >>> is_povm([measurements[:, :, 0, 0], measurements[:, :, 0, 1]])
    True

    Alternatively, the following matrices

    .. math::
        M_0 = \begin{pmatrix}
                                1 & 2 \\
                                3 & 4
                             \end{pmatrix}, \qquad
        M_1 = \begin{pmatrix}
                5 & 6 \\
                7 & 8
              \end{pmatrix}

    does not constitute a POVM set.

    >>> from toqito.measurement_props import is_povm
    >>> import numpy as np
    >>> non_meas_1 = np.array([[1, 2], [3, 4]])
    >>> non_meas_2 = np.array([[5, 6], [7, 8]])
    >>> non_meas = [non_meas_1, non_meas_2]
    >>> is_povm(non_meas)
    False

    References
    ==========
    .. [WikPOVM] Wikipedia: POVM
        https://en.wikipedia.org/wiki/POVM

    :param mat_list: A list of matrices.
    :return: True if set of matrices constitutes a set of measurements, and
             False otherwise.
    """
    dim = mat_list[0].shape[0]

    mat_sum = np.zeros((dim, dim), dtype=complex)
    for mat in mat_list:
        # Each measurement in the set must be positive semidefinite.
        if not is_psd(mat):
            return False
        mat_sum += mat
    # Summing all the measurements from the set must be equal to the identity.
    if not np.allclose(np.identity(dim), mat_sum):
        return False
    return True



"""Tests for dephasing function."""
import unittest
import numpy as np

from toqito.channel_ops import apply_map
from toqito.channels import dephasing


class TestDephasingChannel(unittest.TestCase):

    """Unit test for dephasing."""

    def test_completely_dephasing(self):
        """The completely dephasing channel kills everything off diagonal."""
        test_input_mat = np.array(
            [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
        )

        expected_res = np.array(
            [[1, 0, 0, 0], [0, 6, 0, 0], [0, 0, 11, 0], [0, 0, 0, 16]]
        )

        res = apply_map(test_input_mat, dephasing(4))

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)

    def test_partially_dephasing(self):
        """The partially dephasing channel for `p = 0.5`."""
        test_input_mat = np.array(
            [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
        )

        expected_res = np.array(
            [[17.5, 0, 0, 0], [0, 20, 0, 0], [0, 0, 22.5, 0], [0, 0, 0, 25]]
        )

        res = apply_map(test_input_mat, dephasing(4, 0.5))

        bool_mat = np.isclose(expected_res, res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for cvx_helper function."""
import unittest
import cvxpy
import numpy as np

from toqito.helper import expr_as_np_array, np_array_as_expr


class TestCVXPYHelper(unittest.TestCase):

    """Unit test for cvx_helper."""

    def test_np_array_as_expr(self):
        """Ensure return type is CVX object."""
        test_input_mat = np.array([[1, 4, 7], [2, 5, 8], [3, 6, 9]])

        res_mat = np_array_as_expr(test_input_mat)
        self.assertEqual(isinstance(res_mat, cvxpy.atoms.affine.vstack.Vstack), True)

    def test_expr_as_np_array(self):
        """Ensure return type is numpy object."""
        expr = cvxpy.bmat([[1, 2], [3, 4]])

        res_mat = expr_as_np_array(expr)
        self.assertEqual(isinstance(res_mat, np.ndarray), True)

    def test_expr_as_np_array_scalar(self):
        """Ensure return type is numpy object for scalar expression."""
        cvx_var = cvxpy.Variable()
        self.assertEqual(isinstance(expr_as_np_array(cvx_var), np.ndarray), True)

    def test_expr_as_np_array_vector(self):
        """Ensure return type is numpy object for vector expression."""
        cvx_var = cvxpy.Parameter(5)
        self.assertEqual(isinstance(expr_as_np_array(cvx_var), np.ndarray), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for is_psd function."""
import unittest
import numpy as np

from toqito.matrix_props import is_psd


class TestIsPSD(unittest.TestCase):

    """Unit test for is_psd."""

    def test_is_psd(self):
        """Test that positive semidefinite matrix returns True."""
        mat = np.array([[1, -1], [-1, 1]])
        self.assertEqual(is_psd(mat), True)

    def test_is_not_psd(self):
        """Test that non-positive semidefinite matrix returns False."""
        mat = np.array([[-1, -1], [-1, -1]])
        self.assertEqual(is_psd(mat), False)


if __name__ == "__main__":
    unittest.main()



"""Tests for apply_map function."""
import unittest
import numpy as np

from toqito.channel_ops import apply_map
from toqito.perms import swap_operator


class TestApplyMap(unittest.TestCase):

    """Unit test for apply_map."""

    def test_apply_map_choi(self):
        """
        The swap operator is the Choi matrix of the transpose map.

        The following test is a (non-ideal, but illustrative) way of computing
        the transpose of a matrix.
        """
        test_input_mat = np.array([[1, 4, 7], [2, 5, 8], [3, 6, 9]])

        expected_res = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        res = apply_map(test_input_mat, swap_operator(3))

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_apply_map_kraus(self):
        """
        Apply Kraus map.

        The following test computes PHI(X) where X = [[1, 2], [3, 4]] and
        where PHI is the superoperator defined by:
        Phi(X) = [[1,5],[1,0],[0,2]] X [[0,1][2,3][4,5]].conj().T -
        [[1,0],[0,0],[0,1]] X [[0,0][1,1],[0,0]].conj().T
        """
        test_input_mat = np.array([[1, 2], [3, 4]])

        kraus_1 = np.array([[1, 5], [1, 0], [0, 2]])
        kraus_2 = np.array([[0, 1], [2, 3], [4, 5]])
        kraus_3 = np.array([[-1, 0], [0, 0], [0, -1]])
        kraus_4 = np.array([[0, 0], [1, 1], [0, 0]])

        expected_res = np.array([[22, 95, 174], [2, 8, 14], [8, 29, 64]])

        res = apply_map(test_input_mat, [[kraus_1, kraus_2], [kraus_3, kraus_4]])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for is_povm function."""
import unittest
import numpy as np

from toqito.measurement_props import is_povm
from toqito.random import random_povm


class TestIsPOVM(unittest.TestCase):

    """Unit test for is_povm."""

    def test_is_povm_true(self):
        """Test if valid measurement returns True."""
        dim, num_inputs, num_outputs = 2, 2, 2
        measurements = random_povm(dim, num_inputs, num_outputs)

        self.assertEqual(
            is_povm([measurements[:, :, 0, 0], measurements[:, :, 0, 1]]), True
        )

    def test_is_povm_false(self):
        """Test if invalid POVM returns False."""
        non_meas_1 = np.array([[1, 2], [3, 4]])
        non_meas_2 = np.array([[5, 6], [7, 8]])
        non_meas = [non_meas_1, non_meas_2]

        self.assertEqual(is_povm(non_meas), False)

    def test_is_povm_false_not_sum_identity(self):
        """Test if invalid POVM (does not sum to identity)."""
        non_meas_1 = np.array([[1, 0], [0, 1]])
        non_meas_2 = np.array([[1, 0], [0, 1]])
        non_meas = [non_meas_1, non_meas_2]

        self.assertEqual(is_povm(non_meas), False)


if __name__ == "__main__":
    unittest.main()



"""Tests for update_odometer function."""
import unittest
import numpy as np

from toqito.helper import update_odometer


class TestUpdateOdometer(unittest.TestCase):

    """Unit test for update_odometer."""

    def test_update_odometer_0_0(self):
        """Update odometer from [2, 2] to [0, 0]."""
        vec = np.array([2, 2])
        upper_lim = np.array([3, 2])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([0, 0], res)
        self.assertEqual(np.all(bool_mat), True)

    def test_update_odometer_0_1(self):
        """Update odometer from [0, 0] to [0, 1]."""
        vec = np.array([0, 0])
        upper_lim = np.array([3, 2])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([0, 1], res)
        self.assertEqual(np.all(bool_mat), True)

    def test_update_odometer_1_0(self):
        """Update odometer from [0, 1] to [1, 0]."""
        vec = np.array([0, 1])
        upper_lim = np.array([3, 2])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([1, 0], res)
        self.assertEqual(np.all(bool_mat), True)

    def test_update_odometer_2_0(self):
        """Update odometer from [1, 1] to [2, 0]."""
        vec = np.array([1, 1])
        upper_lim = np.array([3, 2])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([2, 0], res)
        self.assertEqual(np.all(bool_mat), True)

    def test_update_odometer_2_1(self):
        """Update odometer from [2, 0] to [2, 1]."""
        vec = np.array([2, 0])
        upper_lim = np.array([3, 2])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([2, 1], res)
        self.assertEqual(np.all(bool_mat), True)

    def test_update_odometer_2_2(self):
        """Update odometer from [2, 1] to [0, 0]."""
        vec = np.array([2, 1])
        upper_lim = np.array([3, 2])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([0, 0], res)
        self.assertEqual(np.all(bool_mat), True)

    def test_update_odometer_empty(self):
        """Return `None` if empty lists are provided."""
        vec = np.array([])
        upper_lim = np.array([])
        res = update_odometer(vec, upper_lim)

        bool_mat = np.isclose([], res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for ket function."""
import unittest
import numpy as np

from toqito.states import basis


class TestKet(unittest.TestCase):

    """Unit test for ket."""

    def test_ket_0(self):
        """Test for `|0>`."""
        expected_res = np.array([[1], [0]])
        res = basis(2, 0)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_ket_1(self):
        """Test for `|1>`."""
        expected_res = np.array([[0], [1]])
        res = basis(2, 1)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_ket_0000(self):
        """Test for `|0000>`."""
        expected_res = np.array([[1], [0], [0], [0]])
        res = basis(4, 0)

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_invalid_dim(self):
        """Tests for invalid dimension inputs."""
        with self.assertRaises(ValueError):
            basis(4, 4)


if __name__ == "__main__":
    unittest.main()



"""Tests for unambiguous_state_exclusion function."""
import unittest
import numpy as np

from toqito.states import bell
from toqito.state_distinguish import StateDistinguish


class TestUnambiguousStateExclusion(unittest.TestCase):

    """Unit tests for unambiguous_state_exclusion."""

    def test_unambiguous_state_exclusion_one_state(self):
        """Unambiguous state exclusion for single state."""
        mat = bell(0) * bell(0).conj().T
        states = [mat]

        s_d = StateDistinguish(states)
        res = s_d.unambiguous_state_exclusion()
        self.assertEqual(np.isclose(res, 0), True)

    def test_unambiguous_state_exclusion_one_state_vec(self):
        """Unambiguous state exclusion for single vector state."""
        vec = bell(0)
        states = [vec]

        s_d = StateDistinguish(states)
        res = s_d.unambiguous_state_exclusion()
        self.assertEqual(np.isclose(res, 0), True)

    def test_unambiguous_state_exclusion_three_state(self):
        """Unambiguous state exclusion for three Bell state density matrices."""
        mat1 = bell(0) * bell(0).conj().T
        mat2 = bell(1) * bell(1).conj().T
        mat3 = bell(2) * bell(2).conj().T
        states = [mat1, mat2, mat3]
        probs = [1 / 3, 1 / 3, 1 / 3]

        unambiguous_sd = StateDistinguish(states, probs)
        res = unambiguous_sd.unambiguous_state_exclusion()
        self.assertEqual(np.isclose(res, 0), True)

    def test_unambiguous_state_exclusion_three_state_vec(self):
        """Unambiguous state exclusion for three Bell state vectors."""
        mat1 = bell(0)
        mat2 = bell(1)
        mat3 = bell(2)
        states = [mat1, mat2, mat3]
        probs = [1 / 3, 1 / 3, 1 / 3]

        unambiguous_sd = StateDistinguish(states, probs)
        res = unambiguous_sd.unambiguous_state_exclusion()
        self.assertEqual(np.isclose(res, 0), True)

    def test_unambiguous_state_exclusion_complex_three_state_vec(self):
        """Unambiguous state exclusion on complex set of pure states."""
        mat_1 = np.array(
            [
                [0.37166502 + 0.0j, 0.06990262 + 0.00381928j, 0.44548935 - 0.17369055j],
                [0.06990262 - 0.00381928j, 0.01318651 + 0.0j, 0.0820026 - 0.03724556j],
                [0.44548935 + 0.17369055j, 0.0820026 + 0.03724556j, 0.61514848 + 0.0j],
            ]
        )

        mat_2 = np.array(
            [
                [0.03351844 + 0.0j, 0.08195346 + 0.02701392j, 0.15185457 + 0.0434629j],
                [0.08195346 - 0.02701392j, 0.22215 + 0.0j, 0.40631695 - 0.01611805j],
                [0.15185457 - 0.0434629j, 0.40631695 + 0.01611805j, 0.74433156 + 0.0j],
            ]
        )

        mat_3 = np.array(
            [
                [0.51449115 + 0.0j, 0.23369567 + 0.15751255j, 0.41173315 - 0.02901644j],
                [0.23369567 - 0.15751255j, 0.15437363 + 0.0j, 0.17813679 - 0.13923301j],
                [0.41173315 + 0.02901644j, 0.17813679 + 0.13923301j, 0.33113522 + 0.0j],
            ]
        )
        states = [mat_1, mat_2, mat_3]
        probs = [1 / 3, 1 / 3, 1 / 3]

        unambiguous_sd = StateDistinguish(states, probs)
        res = unambiguous_sd.unambiguous_state_exclusion()
        self.assertEqual(np.isclose(res, 0, atol=1e-06), True)


if __name__ == "__main__":
    unittest.main()



"""Tests for permutation_operator function."""
import unittest
import numpy as np

from toqito.perms import permutation_operator


class TestPermutationOperator(unittest.TestCase):

    """Unit test for permutation_operator."""

    def test_standard_swap(self):
        """Generates the standard swap operator on two qubits."""
        expected_res = np.array(
            [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]
        )

        res = permutation_operator(2, [2, 1])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_standard_swap_list_dim(self):
        """Generates the standard swap operator on two qubits."""
        expected_res = np.array(
            [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]]
        )

        res = permutation_operator([2, 2], [2, 1])

        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_sparse_option(self):
        """Sparse swap operator on two qutrits."""
        res = permutation_operator(3, [2, 1], False, True)

        self.assertEqual(res[0][0], 1)


if __name__ == "__main__":
    unittest.main()



"""Tests for hadamard function."""
import unittest
import numpy as np

from toqito.matrices import hadamard


class TestHadamard(unittest.TestCase):

    """Unit test for hadamard."""

    def test_hadamard_0(self):
        """Test for Hadamard function when n = 0."""
        res = hadamard(0)
        expected_res = np.array([[1]])
        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_hadamard_1(self):
        """Test for Hadamard function when n = 1."""
        res = hadamard(1)
        expected_res = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_hadamard_2(self):
        """Test for Hadamard function when n = 2."""
        res = hadamard(2)
        expected_res = (
            1
            / 2
            * np.array([[1, 1, 1, 1], [1, -1, 1, -1], [1, 1, -1, -1], [1, -1, -1, 1]])
        )
        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_hadamard_3(self):
        """Test for Hadamard function when n = 3."""
        res = hadamard(3)
        expected_res = (
            1
            / (2 ** (3 / 2))
            * np.array(
                [
                    [1, 1, 1, 1, 1, 1, 1, 1],
                    [1, -1, 1, -1, 1, -1, 1, -1],
                    [1, 1, -1, -1, 1, 1, -1, -1],
                    [1, -1, -1, 1, 1, -1, -1, 1],
                    [1, 1, 1, 1, -1, -1, -1, -1],
                    [1, -1, 1, -1, -1, 1, -1, 1],
                    [1, 1, -1, -1, -1, -1, 1, 1],
                    [1, -1, -1, 1, -1, 1, 1, -1],
                ]
            )
        )
        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)

    def test_hadamard_negative(self):
        """Input must be non-negative."""
        with self.assertRaises(ValueError):
            hadamard(-1)


if __name__ == "__main__":
    unittest.main()



"""Tests for state_distinguishability function."""
import unittest
import numpy as np

from toqito.states import basis, bell
from toqito.state_distinguish import StateDistinguish


class TestStateDistinguishability(unittest.TestCase):

    """Unit tests for state_distinguishability."""

    def test_state_distinguishability_one_state(self):
        """State distinguishability for single state."""
        rho = bell(0) * bell(0).conj().T
        states = [rho]

        s_d = StateDistinguish(states)
        res = s_d.state_distinguishability()
        self.assertEqual(np.isclose(res, 1), True)

    def test_state_distinguishability_one_state_vec(self):
        """State distinguishability for single vector state."""
        rho = bell(0)
        states = [rho]

        s_d = StateDistinguish(states)
        res = s_d.state_distinguishability()
        self.assertEqual(np.isclose(res, 1), True)

    def test_state_distinguishability_two_states(self):
        """State distinguishability for two state density matrices."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        e_00 = e_0 * e_0.conj().T
        e_11 = e_1 * e_1.conj().T
        states = [e_00, e_11]
        probs = [1 / 2, 1 / 2]

        s_d = StateDistinguish(states, probs)
        res = s_d.state_distinguishability()
        self.assertEqual(np.isclose(res, 1 / 2), True)

    def test_state_distinguishability_three_state_vec(self):
        """State distinguishability for two state vectors."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        states = [e_0, e_1]
        probs = [1 / 2, 1 / 2]

        s_d = StateDistinguish(states, probs)
        res = s_d.state_distinguishability()
        self.assertEqual(np.isclose(res, 1 / 2), True)

    def test_invalid_state_distinguishability_probs(self):
        """Invalid probability vector for state distinguishability."""
        with self.assertRaises(ValueError):
            rho1 = bell(0) * bell(0).conj().T
            rho2 = bell(1) * bell(1).conj().T
            states = [rho1, rho2]
            s_d = StateDistinguish(states, [1, 2, 3])
            s_d.state_distinguishability()

    def test_invalid_state_distinguishability_states(self):
        """Invalid number of states for state distinguishability."""
        with self.assertRaises(ValueError):
            states = []
            s_d = StateDistinguish(states)
            s_d.state_distinguishability()


if __name__ == "__main__":
    unittest.main()



"""Tests for cnot function."""
import unittest
import numpy as np

from toqito.matrices import cnot


class TestCNOT(unittest.TestCase):

    """Unit test for cnot."""

    def test_cnot(self):
        """Test standard CNOT gate."""
        res = cnot()
        expected_res = np.array(
            [[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]
        )
        bool_mat = np.isclose(res, expected_res)
        self.assertEqual(np.all(bool_mat), True)


if __name__ == "__main__":
    unittest.main()
